/*++++++++++++++
.IDENTIFICATION tabmap.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  4.0   01-Apr-2002 New package
.VERSION  4.1   18-Dec-2002: Prepared as a full package
.VERSION  4.2   01-Feb-2003: Accept $1 .. $n as column names
.VERSION  4.3   20-Aug-2003: BUG in case of NaN
.VERSION  4.4   16-Feb-2004: Accept /./ for subcatalog
.VERSION  4.5   30-Dec-2004: Accept -asci, -tex, etc...
.VERSION  4.6   06-Sep-2005: Option -n (stop after n lines)
.VERSION  4.7   22-Oct-2005: Option -b (remove blanks), -head [u-1]
.VERSION  4.8   11-Jul-2006: Accept ${} or $0 (whole record)
.VERSION  4.81  01-Jan-2007: Right-alignment of label
.VERSION  4.82  18-Mar-2007: error message if file not found
.VERSION  4.83  02-Jun-2007: Cosmetics (-Wall)
.VERSION  4.9   16-Feb-2008: Accept -e (idem -map), @ (instead of $)
.VERSION  4.91  25-May-2008: Comment lines may be longer than lrecl
.VERSION  4.92  11-Jul-2008: Leave the \ . -keep_cols options.
.VERSION  5.0   09-Sep-2008: Accept subelements
.VERSION  5.1   21-Sep-2008: Accept -skip -ml
.VERSION  5.2   30-Dec-2008: +n option (tail)
.VERSION  5.3   18-Feb-2009: vertical presentation when a number in -form
.VERSION  5.4   21-Feb-2009: Accept . for all files
.VERSION  5.5   30-Apr-2009: Pb with relative names. HTML output changed.
.VERSION  5.51  11-May-2009: Pb with selected files
.VERSION  5.6   21-Sep-2009: No need to specify full complex values
.VERSION  5.61  29-Oct-2009: No error message 'Strange mode'
.VERSION  5.62  17-Nov-2009: +n (number of lines to skip) starts from 1
.VERSION  5.63  15-Dec-2009: 64-bit
.VERSION  5.64  09-Jun-2010: pb +n
.VERSION  5.65  14-Jul-2010: pb \n in title of table
.VERSION  5.66  01-Dec-2010: Bug last column !!
.VERSION  5.67  13-Jan-2011: Don't edit the --- columns...
.VERSION  5.68  16-Mar-2011: With -t argument ==> use ReadMe from -t
.VERSION  5.69  24-Mar-2011: Search for ReadMe hierarchies
.VERSION  5.7   24-Oct-2011: -i: interprets escape chars like \n
.VERSION  5.71  13-Mar-2012: test specified null value
.VERSION  5.72  07-May-2012: Accept '#' for record number in list
.VERSION  5.73  09-Jun-2012: Label corrected in parfile output (e.g. was l_mag)
.VERSION  5.74  23-Jun-2012: Option -r (specified the ReadMe file)
.VERSION  5.75  24-Oct-2012: Pb with complex values
.VERSION  5.8   24-Nov-2012: Accept @{@pos}, @{@PK}
.VERSION  5.81  22-Mar-2013: 64-bit
.VERSION  5.82  26-Dec-2013: Skip the prefix in par_echo
.VERSION  5.9   02-Jan-2014: Accept "start" as skip+1
			Use the "gzi" index if existing
.VERSION  5.91  06-Mar-2014: -i option: add newline in par mode
.VERSION  5.92  10-Apr-2014: Full file name in TABMAP structure
.VERSION  5.93  24-Jun-2014: accept -ml -1 for all lines
.VERSION  5.94  22-Jul-2014: accept files separated by ,
.VERSION  5.95  19-Oct-2014: bug fixed 
.COMMENTS       Deal with TABMAP definitions and structures
	The "mapping" consists in extracting a list of columns
	from a table and gluing with other pieces of text.
	The syntax of the definition knows
	- the '#' indicates a comment until the end-of-line, but
	  #!Ep=(epoch)  #!Eq=(equinox)
	  #!labw=(width of label, negative for left adjustment)
	- quoted strings (surrounded by ' or ") with C-like definitions
	- the , (indicates a separator)
	- the ; (indicates a newline)
	- strings starting with \ with C-like conventions
	- label names
    Special formats: 'X' = new column, 'W' = whole line (record)
---------------*/

#define VERSION "5.95 (2014-10-19)"

#ifndef DEBUG
#define DEBUG 0
#endif
#ifndef TEST
#define TEST 0
#endif

#include <catstd.h>
#include <filutil.h>
#include <buffer.h>

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ABS(x)		((x) < 0 ? -(x) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define bzero(a,n)		memset(a, 0, n)
#define RULE(mode)	rule(mode == 'H' ? '_' : '-')

#if DEBUG
static char inbyte;
#endif

typedef char *PCHAR;		/* Use typedef to declare a function
				   returning a pointer to char */
typedef PCHAR (*PCHAR_FCT)();	/* Function returning pointer to char */
typedef int   (*INT_FCT)();	/* Function returning an integer .... */

static char modes[] = "acpthH" ;
	/* Acceptable modes: ascii - CSV - par - tex - htx - html */

static int nowrap = 16 ;	/* Limit for NoWrap  */
static char omapc = 0;		/* -mapc */
static char *readme_file = "ReadMe";

/* Printing options */
#define P_note0 0x8000		/* Print notes in header  */
#define P_note1 0x4000		/* Print notes after data */
static int P_opt = 0;

/* All variables / functions are in the same order   a c p t h H */

/* Comment strings */
static char com0p[2] = "%";
static char *com0[] = { "#", "#", com0p, "%", "%", "<!-- " };
static char *com1[] = { "" , "" ,   "" , "" , "" , " -->"  };
/*                       0    1     2     3    4     5	*/

	/* The list of well-known Identifiers */
static char ids[] = \
    "NGC,IC,Variable,SAO,PPM,HR,HD,HDE,BD,CD,CPD,Seq,PKS,PN,HIP,TYC,GSC," ;

static char sepchar[] = " ";	/* Separator of columns */

static int exit_status = 0;
static int max_lines = 0;
static int oskip = 0;
static int otail = 0;
static int fop = FOP_quiet;	/* Options to process 	*/
#define verbop (fop&FOP_v)
static char keep_cols = 0;	/* Option -keep_cols    */
static char issue_titles = 1;	/* Change by -data_only	*/
static char tourop = 0;		/* -tour option		*/
static char bopt   = 0;		/* -b    option		*/
static char *null_value;	/* How to convert NULLs	*/
static char *mapping;		/* -map    instructions	*/
static char *form;		/* -form   contents	*/
static char *tableform=0;	/* -t option            */
static char head_def[] = "H1u-";
static char *head_opt=head_def;	/* Header, label, unit  */

static CATINFO *theCat ;	/* Set by main, remove once edited */
static FILINFO *theFile ;	/* Set by tbm_new       */
static COLINFO  Wrecord;	/* For the whole record */
static COLINFO  Crecord;	/* For the record number */
static int recno;
static char arec[20];	/* Edited recno */

/* Meaning of number of stars (for trimming) */
static char flagstars[] = { F_hMAP|F_tMAP, F_tMAP, F_hMAP|F_tMAP|F_sMAP } ;

/* Special TeX / HTML values */
static char *texspec[] = {
   "%\\%", "{\\{", "}\\}", "_\\_", "$\\$", "^\\^{ }", "#\\#", "\\$\\backslash$",
   0 };
static char *htmspec[] = {
   "&&amp;", "<&lt;", ">&gt;", "\"&quot;", "'&apos;", 0 };

static char **aspecs[] = {	/* List of Special Characters */
    (char **)0, 	/* 'a' */
    (char **)0,		/* 'c' */
    (char **)0, 	/* 'p' */
    texspec,		/* 't' */
    texspec,		/* 'h' */
    htmspec,		/* 'H' */
} ;

static char *rules[] = {	/* How to write a rule */
    (char *)0 		/* 'a' */,
    (char *)0		/* 'c' */,
    "" 			/* 'p' */,
    "{\\hline}\n"	/* 't' */,
    "{\\hline}\n"	/* 'h' */,
    "<HR>\n"		/* 'H' */,
} ;

static char *fmsg[] = {	/* How to write a Message */
    "#***%s\n"		/* 'a' */,
    "\n#***%s\n"	/* 'c' */,
    "\n%***%s\n" 	/* 'p' */,
    "\\\\\n{\\bf %s}\n" /* 't' */,
    "{\\br}{\\bf\\fg{red}%s}\n" /* 'h' */,
    "\n<font color='#FF0000'><b>%s</b></font>\n" /* 'H' */,
} ;

static char *dochead[] = {   /* Document Head */
    "## (from tabmap V%s) %s\n", /*a*/
    "## (from tabmap V%s) %s\n", /*c*/
    "%%%% (from tabmap V%s) %s\n", /*p*/
    "\\begin{document} % tabmap-V%s  %s\n"	/* 't' */,
    "%%%% (from tabmap)"	/* 'h' */,
  "<!-- from tabmap V%s %s -->\n\
<LINK rel='stylesheet' type='text/css' href='/vizier/vizier.css'>\n\
<LINK rel='stylesheet' type='text/css' \
href='/vizier/lib/sortable/tablesort.css'>\n",
};
static char *doctail[] = {   /* Document Tail, argument = table/catalog */
    "", /*a*/
    "", /*c*/
    "%% (from tabmap)", /*p*/
    "\\end{document}\n"	/* 't' */,
    "%% (End)"	/* 'h' */,
    "<HR NOSHADE>\n",
    /*
"<div id='CDStail' class='CDSbgcol4 CDScol3'>\n\
<style type='text/css'>\n\
#CDStail { padding-left:9.5em; text-align: right; }\n\
#CDStail img{ vertical-align: middle; border:0px; }\n\
</style>\n\
<span style='font-size: smaller'>&copy;ULP/CNRS</span>\n\
&nbsp;Contact:&nbsp;<a href='\
mailto:question@simbad.u-strasbg.fr?Subject=%s'><img src=\
'http://simbad.u-strasbg.fr/Icons/envelop1.gif' align='middle' \
border='0' alt='Question@simbad'></a>\n\
</div>\n", */
};
static char *conclude;

static char *cat_titles[] = { 	/* How to write catalog title */
    /* Arguments are code title */
    /*a*/"\
#---------------------------------------------------------------------------\n\
#-- %s %s\n\
#--------------------------------------------------------------------------\n",
    /*c*/"\
#--------------------------------------------------------------------------\n\
#-- %s %s\n\
#--------------------------------------------------------------------------\n",
    /*p*/"\
%%--------------------------------------------------------------------------\n\
%%-- %s %s\n\
%%--------------------------------------------------------------------------\n",
    /*t*/"\\section*{\\underline{%s} %s}\n",
    /*h*/"\\section*{{\\fg{blue4}%s} %s}\n",
    /*H*/"<H3><font color='#00008b'>%s</font> %s</H3>\n"
};
static char *tab_titles[] = { 	/* How to write table title */
  /* Arguments are code title records */
  /*a*/"#---Table: %s %s (%d records)\n"   ,
  /*c*/"#---Table: %s %s (%d records)\n"   ,
  /*p*/"%%---Table: %s %s (%d records)\n" ,
  /*t*/"\\subsection*{\\underline{%s} %s {\\em(%d records)}}\n"	,
  "\\subsection*{{\\fg{blue4}\\plain{%s}} %s {\\em\\fg{blue4}(%d records)}}\n",
  "<H4><font color='#00008b'>%s</font> %s \
   <i>(<font color='#00008b'>%d records</i></font>)</H4>\n"
};
static char TRstyle[] = "<TR class='tuple-2'>";
static int  TRmod;     /* Modulo value */
static int  iTRmod;    /* Index of class in TRstyle */

static char blanks[] =
"                                                                          ";

static char usage[] = "\
Usage: tabmap [-v] [-b] [-head {1u-}] [-data_only] [-notes] [+notes] \n\
       [-null text] -form type [-keep] [-map[c] instruction] [-ml max_lines]\n\
       [-sl nr|-start recno] [+n] [-tour] [-i text] [-r ReadMe] [-t table] \n\
       [file[,file]|-]... [-go other_tables...]\n\
";

static char help[] = "\
Transform the file described by a ReadMe into other formats (-form option):\n\
   a = -ascii (aligned columns)\n\
   c = -csv (tab-separated)\n\
   p = -parfile; options in this form:\n\
        S = recognize SIMBAD name\n\
        0 = don't edit the 'null' columns (those having no value)\n\
        . = align labels, values, units, and explain each line\n\
       ^c = use character 'c' as leftmost char (default is %)\n\
       (w,v,u,x) = assign a length for label, value, unit, explanation\n\
                  (aligned edition), missing values align everything.\n\
   t = -latex\n\
   h = -htx\n\
   H = -html\n\
The -form option may include | || (add column separators), or * for trim.\n\
\n\
The mapping instructions (-map) are written as ${colname}, ${#col_number} or\n\
just $n (equivalent to ${#n}); ${#} represents the column number (from 1),\n\
and the whole line by ${} or $0. Trimming supported operations are:\n\
  ${*colname} -- ${**colname} = tail trim, ${***colname} = squeezed trim\n\
Control characters are written e.g. ${\\n} for the newline; however \\n is\n\
understood as the newline with the -mapc (instead of -map) option.\n\
\n\
Other options are:\n\
     -v: verbose option\n\
     -b: remove the useless blanks (in CSV or HTML)\n\
  -head: the header consists of label (1) unit (u) dashes (-)\n\
  -data: write the data only (by default the column headers are added)\n\
  -note: write out the notes related to the table(s)\n\
  +note: write out the notes after the data table(s)\n\
 -i txt: just issue the specified text (e.g. between 2 files)\n\
  -keep: keep original columns (default is to merge e.g. RAh RAm RAs)\n\
  -ml #: maximal number of lines to write out (-n e.g. -10 also acceptable)\n\
  -null: how to write the NULL numeric values (default is input value)\n\
         (value may be attached if -null=text is used instead of -null text)\n\
  -sl #: number of data lines lines (records) to skip from input\n\
        [alternatively -st # indicates the record# (from 1) to start]\n\
    -r: specify the name of the 'ReadMe' file, if not \"ReadMe\"\n\
    -t: specify a \"tabular template\" (input file structured like table)\n\
   -go: edit what's asked, and start a new run (edition of other table(s)\n\
 -tour: show the various possibilities, editing 10 lines in different formats\n\
    +n: which line number to start (from 1, like tail(1))\n\
\n\
 file  may include the directory in which the files are to be found;\n\
       for files residing in subdirectories, use the \"./\" convention to \n\
       indicate the ReadMe directory, as e.g. III/75/./sp/hd123.dat\n\
       A comma-separated list of files is treated as a single file,\n\
       while a blank line separates each file in a list of files.\n\
 If file is '.' (just a dot), all files making up the catalog are edited.\n\
";

/*==================================================================
		Interpretation of special fields
 *==================================================================*/

static int interpret_escape(char *text, int omapc)
/*++++++++++++++++
.PURPOSE  Replace in text the \ characters.
.RETURNS  The last char read
.REMARKS  omapc!=0 ==> interpret escape chars
-----------------*/
{
  char *p, *s, c=0; int i;
  int inbrace = 0;

    for (p=s=text; *s; s++, p++) {
	c = *s;
        if ((c == '\\') && omapc) {
            i = str1(s, p);	/* str1 write the interpretation */
	    if (i > 0) --i;
	    s += i;
	    continue;
        }
	else if (c == '@') {
	    if ((inbrace) || (s[1] == '{' /*'}'*/))
		c = '$';
	}
	else if (c == '{') inbrace++;
	else if (c == '}') inbrace--;
        *p = c;
    }
    *p = 0;
    return(c);
}

static void BUF_AppendChars(BUFFER *b, int c, int count)
/*++++++++++++++++
.PURPOSE  Append the character -- repeated 'count'
.RETURNS  ---
-----------------*/
  /* MOD: b = Buffer concerned 	*/
  /* IN: text = Words to append	*/
{
  static char buf[2] ;
    buf[0] = c ;
    while (--count >= 0) {
	if (b->used < b->allocated) b->buf[b->used++] = c;
        else BUF_AppendItem(b, char, buf) ;
    }
}

static int buf_text(char *text, BUFFER *b, int imode)
/*++++++++++++++++
.PURPOSE  Print out a text -- escape the problematic chars
.RETURNS  Number of bytes
.REMARKS  imode = index in com0 / com1
-----------------*/
{
  char **aspec, *p, c ;
  int i;
    aspec = aspecs[imode] ;
    for (p=text; *p; p++) {
	c = *p ;
	if (aspec) {
#if 0
	    if ((aspec == htmspec) && (c == '-')) { /* dashes */
		if (p[1] == '-') {
		    if (p[2] == '-') {
			BUF_AppendString(b, "&#2014;") ;  /* mdash */
			p += 2;
		    }
		    else {
			BUF_AppendString(b, "&#2013;");   /* ndash */
			p++;
		    }
		}
		else {
		    BUF_AppendString(b, "&#2012;");       /* fig. dash */
		}
	    }
	    else 
#endif
	    {
	        for (i=0; aspec[i] && *aspec[i] != c; i++) ;
	        if (aspec[i])  BUF_AppendString(b, aspec[i]+1);
	        else BUF_AppendChars(b, c, 1);
	    }
	}
	else BUF_AppendChars(b, c, 1);
    }
    BUF_AppendString(b, ""); 	/* Terminate with NULL */
    return(p-text) ;
}

static int pr_com(char *text, FILE *fout, int imode)
/*++++++++++++++++
.PURPOSE  Print out lines of comments
.RETURNS  Number of lines written out
.REMARKS  imode = index in com0 / com1
-----------------*/
{
  char **aspec, *p, c ;
  int i, n = 0;

    aspec = aspecs[imode] ;
    /* if (*com0[imode]) fprintf(fout, com0[imode]) ; */
    for (p=text, c='\n'; *p; p++) {
	if (c == '\n') {
	    n++;
	    if (!*com1[imode]) fprintf(fout, "%s", com0[imode]) ;
	}
	c = *p ;
	if ((aspec == htmspec) && (c == '-') && (p[1] == '-')) c = '_' ;
	else if (aspec) {
	    for (i=0; aspec[i] && *aspec[i] != c; i++) ;
	    if (aspec[i])  c=0, fprintf(fout, "%s", aspec[i]+1);
	}
	if (c) fputc(c, fout) ;
    }
    /* fprintf(fout, "%s", com1[imode]) ; */
    /* if (c != '\n') fputc('\n', fout) ; */
    return(n) ;
}

static char *thePos(TABINFO *T)
/*++++++++++++++++
.PURPOSE  Get the components of a position 
.RETURNS  The developped meaning, e.g. ${RAh}:${RAm}:${RAs}${DE-}... / NULL
.REMARKS  Static buffer
-----------------*/
{
  static char result[256], *t;
  COLINFO *col;
  int i, m=sizeof(result)-8, r=0, sign=0;
    if(T->posfmt[0]<0) return((char *)0); /* No position */
    for (i=0; (i<ITEMS(T->posfmt)) && T->posfmt[i]>=0 && (r<m); i++) {
	col = T->afmt+T->posfmt[i];
	if (col->flags&F_ACCU) { if(!sign) result[r++] = ':'; }
	result[r++] = '$'; result[r++] = '{';
	for (t=col->label; *t && (r<m); t++) result[r++] = *t;
	result[r++] = '}';
	sign = col->flags&F_SIGN;
    }
    result[r] = 0;
    return(result);
}

static char *thePK(TABINFO *T)
/*++++++++++++++++
.PURPOSE  Get the components of a primary key
.RETURNS  The developped meaning, e.g. ${RAh}:${RAm}:${RAs}${DE-}... / NULL
.REMARKS  Static buffer
-----------------*/
{
  static char result[256], *t;
  COLINFO *col;
  int i, n, r=0, m=sizeof(result)-8;
     n = T->nfmt;
    for (i=0; (i<n) && (r<m); i++) {
	col = T->afmt+i;
	if (!(col->flags&F_isPK)) continue;
	if(r>0) result[r++] = ':';
	result[r++] = '$'; result[r++] = '{';
	for (t=col->label; *t && (r<m); t++) result[r++] = *t;
	result[r++] = '}';
    }
    result[r] = 0;
    return(r ? result : (char *)0);
}

static COLINFO *theCol(TABINFO *T, char *text, char *flags, int flagstar)
/*++++++++++++++++
.PURPOSE  Interpret what's in ${...}
.RETURNS  The COLINFO -- 'X' if a new one
.REMARKS  The flags are set in
-----------------*/
  /* IN: tab = FILINFO or TABINFO value */
  /* IN: text = Text to interpret...	*/
  /* OUT: flags = Flags head/tail strip	*/
  /* IN: flagstar = 1 if star is a flag, F_vMAP is a simple text */
{
  static COLINFO aCol ;
  static char junk[2] ;
  COLINFO *c;
  char *t=(char *)0;
  int n;

    if (!aCol.ff) {
        aCol.ff = 'X';
	aCol.rep=1;
	aCol.signature = cad_signature("\3MAP");
    }
    aCol.value = junk;
    aCol.oid   = cad_oid();
    aCol.len = 1;
    *flags = 0 ;
    if (flagstar & F_vMAP) {
        *flags = F_vMAP ;
	goto CopyText ;
    }

    /* Count * = trim  **=TAIL trim, ***=Squeeze trim */
    if (flagstar && (text[0] == '*')) {
        for (n=1; text[n] == '*'; n++) ;
	if (text[n]){
	    text += n;
	    n = MIN(n, sizeof(flagstars)) ;
	    *flags = flagstars[n-1] ;
	}
	else {		/* e.g. ${**} ==> write ** */
	    *flags |= F_vMAP ;
	    flagstar = 0 ;
	}
    }

    if (*text == '$') {
        if (strcmp(text, "$table") == 0) {
	    if (theFile) {
		aCol.value=theFile->dfile;
		aCol.len = strlen(theFile->dfile) ;
	        return (&aCol) ;
	    }
	    fprintf(stderr, "#+++${$table} requires FILINFO\n") ;
	    *flags |= F_vMAP ;
	}
    }

    else if ((*text == '#') && isdigit(text[1])) {
	n = atoi(text+1);
	for (t=text+1; isdigit(*t); t++) ;
	if ((n > 0) && (n <= T->nfmt) && (!*t)) /* Effectively e.g. $10 */
	   return(T->afmt + n - 1);
    }

    else if (*text == '\\') {
        n = str1(text, junk);
	if (!text[n]) {	/* Interpretation is OK */
	    *flags |= F_vMAP ;
	    junk[1] = 0;
	    text = junk;
	}
    }

    else { char *adollar;			/* Mod V5.0 */
	if ((adollar = strchr(text, '$'))) {	/* Sub-element ? */
	    *adollar = 0;
	    c = tbf_findlabel(T, text);
	    *adollar = '$';
	}
	else c = tbf_findlabel(T, text);
	if (c) {	/* A column name was found */
	    if (adollar) *flags |= F_XMAP;
	    return(c) ;
	}
    }

  CopyText:
    if (*flags & F_vMAP) {
	aCol.value = text;
        aCol.len = strlen(text) ;
	return (&aCol) ;
    }

    fprintf(stderr, "#+++Can't interpret ${%s}\n", text) ;
    return((COLINFO *)0) ;
}

/*==================================================================
		Creator
 *==================================================================*/
typedef char *PTR;
typedef PTR (*PTR_FCT)() ;      /* Function returning char pointer */
static struct {
    char *name; PTR_FCT prog; 
} specials[] = {
    { "${$pos}", thePos },
    { "${$PK}",  thePK  },
};

static TABMAP *new(TABINFO *T, char *text)
/*++++++++++++++++
.PURPOSE  Interpret the "instruction" text
.RETURNS  The interpreted text / NULL
.REMARKS  The whole file is loaded in a single run.
-----------------*/
  /* IN: tab = FILINFO or TABINFO structure	*/
  /* IN: text = The "mapping" instructions	*/
{
  COLINFO *pf, *px, **af;
  TABMAP *map;
  char *p, *t, *s, *flags, *tofree=0, mode, junk[2], isspec[256];
  int  i, n, nf, ls, lspec, errors, level, nsubs=0;
  int globop=0 ;		/* Global Option */
  int colno=0;
  int bytes;

    /* Converts the ${$pos} (V5.75)*/
    t = text;
    while((t=strstr(t, "${$"))) {
	i = (t-text);
	p = 0;
	for (n=0; n<ITEMS(specials); n++) {
	    if (strncmp(t, specials[n].name, strlen(specials[n].name)) == 0) {
		p = (*specials[n].prog)(T);
		break;
	    }
	}
	if (!p) { t += 3; continue; }
	/* Substitute ${$pos} by the interpretation */
	s = malloc(strlen(text)+strlen(p));
	strncpy(s, text, i); strcpy(s+i, p); strcat(s+i, text+i+7);
	if (tofree) free(tofree); 
	tofree = text = s; t = text+i+3;
    }
    /* Initialization: first tells the first valid instruction */
    n = nf = ls = lspec = level = errors = 0;
    t = text;
    mode = 0;
    bzero(isspec, sizeof(isspec)) ;
    isspec[0] = 1; isspec['\n'] = 1; isspec['$'] = 1; isspec['#'] = 1;
    isspec['!'] = 1;

    /* Count the number of elements in the list (n),
       the number of existing columns (nf),
       and the global size of the strings (ls)
    */
    p = (char *)0;
    while(*t) {
    	if ((colno==0) && ((*t=='#')||(*t=='!'))) {	/* Comment */
    	    i = strloc(t, '\n'); junk[0] = t[i]; t[i] = 0;
	    if(strncmp(t, "#!form=", 7) == 0) { char *f;
	        mode = t[7] ;
		if (!strchr(modes, mode)) {
		    fprintf(stderr, "****Non-interpretable %s\n", t) ;
		    errors++ ;
		}
		/* Ignore NULL columns: indicated by a '0' (zero),
		 * before any additional spec (within parentheses) */
		for (f=t+7; isalpha(*f); f++) ;
		while (isgraph(*f)) {
		    if (*f == '*') ;
		    else if (*f == '0') globop |= F_0MAP;
		    else if (*f == '.') globop |= F_0xMAP;
		    else if (*f == '(') f += strloc(f, ')');
	            else if (*f == '^') com0p[0] = *++f;
	            else if (*f == '|') ;
		    else if ((mode == 'c') && ispunct(*f)) ;
		    else fprintf(stderr,
			    "#+++Strange mode '%c', ignored in %s\n",
			    *f, t);
		    if (*f) f++;
		}
	    }
    	    t += i; *t = junk[0];
	    if (*t) t++ ;
    	    continue;
	}
	if ((*t == '$') && (t[1] == '{')) {
	    t += 2 ;
	    i = strloc(t, '}') ;
	    if (i == 0) {	/* V4.8: ${} = whole line */
		n++;
		ls += Wrecord.len;	/* V5.95 */
		goto next1;
	    }
	    if ((i == 1) && (t[0] == '#')) {	/* Counter */
		n++;
		ls += Crecord.len;	/* V5.95 */
		goto next1;
	    }
	    /* V5.72: "recno" == # */
	    if ((i==5) && (strncmp(t, "recno", 5) == 0)) {
		n++;
		ls += Crecord.len;	/* V5.95 */
		goto next1;
	    }
	    junk[0] = t[i]; t[i] = 0 ;
	    if ((pf = theCol(T, t, junk+1, 1))) {
		n++ ;
	        if (pf->ff == 'X') ls += pf->len;
		else nf++ ;
		if (junk[1]&F_XMAP) nsubs++;
	    }
	    else errors++ ;
	    t[i] = junk[0];
	  next1:
	    t += i ;
	    if (*t) t++ ;
	    continue;
	}

	else if ((*t == '$') && isdigit(t[1])) {
	    t[0] = '#';
	    for (i=1; isdigit(t[i]); i++) ;
	    junk[0] = t[i]; t[i] = 0;
	    if (atoi(t+1) == 0)	{ /* V4.8: $0 = whole line */
		   pf = &Wrecord; ls += Wrecord.len; }	/* V5.95 */
	    else { pf = theCol(T, t, junk+1, 1); nf++; }
	    t[0] = '$'; t[i] = junk[0];
	    if (pf) { n++; t += i; continue; }
	}
	else if ((*t == '$') && (t[1] == '#')) {	/* Counter */
	    n++;
	    ls += Crecord.len;	/* V5.95 */
	    t += 2;
	    continue;
	}

	/* Simple text, until next comment / end of line */
	for (p=t; isspec[(*p)&255] == 0; p++) ;
	i =  p-t; if (i == 0) i=1;
	if (t[i] == '\n') { ++i ; colno=0; }
	else colno += i;
	n++ ;
	ls += i ;
	t += i ;
    }

    if (!mode) mode='a', fprintf(stderr,
        "#+++tbm_new: 'form' not specified, ascii assumed") ;
    if (!strchr(modes, mode)) {
        fprintf(stderr, "****tbm_new: unknown mode '%c'\n", mode) ;
	if (tofree) free(tofree);
	return((TABMAP *)0) ;
    }
    if (errors) {
	if (tofree) free(tofree);
	return((TABMAP *)0) ;
    }

    /* Allocate now the space for the descriptors:
    	don't forget the array of COLINFO' pointers, the new
    	set of 'X' COLINFO for in-line text, the end-of-line text
    	and the various strings
    */
    i = n - nf;			/* Number of suplementary 'X' COLINFO 	*/
    bytes = sizeof(TABMAP)
    	+ n*sizeof(COLINFO *)		/* Array of COLINFO pointers 	*/
	+ n				/* Array of aflags		*/
    	+ i*sizeof(COLINFO)		/* Set of 'X' COLINFO struct.	*/
	+ nsubs*sizeof(int)		/* Elements with substructures  */
    	+ ls 				/* Space required for strings	*/
	+ i				/* String NUL terminators	*/
	+ 1 + strlen(text);
    map = (TABMAP *)malloc(bytes);
    map->signature = cad_signature("\3MAP") ;
    map->oid = cad_oid();
    map->tabform = T;
    map->next = T->mapping ; T->mapping = map ;
    map->dfile = (char *)0;
    map->mode = mode ;
    map->bars = 0 ;
    map->flags= globop ;
    map->prefix = map->postfix = (char *)0 ;
    if (theFile)
         map->equinox = theFile->equinox, map->epoch = theFile->epoch ;
    else map->equinox = T->equinox,       map->epoch = T->epoch ;
    map->nfmt = n ;
    map->afmt = af = (COLINFO **)(map+1);
    px = (COLINFO *)(af+n);		/* New 'X' Formats		*/
    bzero(px, i*sizeof(COLINFO));
    map->asubs  = (int *)(px+i) ;
    map->aflags = flags = (char *)(map->asubs+nsubs);
    s = map->aflags + n ;
    map->text   = s ; strcpy(s, text); s += strlen(s)+1 ;

    t = text;
    nsubs = colno = 0;
    while(*t) {
    	if ((colno==0) && ((*t=='#')||(*t=='!'))) {	/* Comment */
    	    i = strloc(t, '\n');
    	    t += i;
	    if (*t) t++ ;
    	    continue;
	}
	if ((*t == '$') && (t[1] == '{')) {
	    t += 2 ;
	    p = t ;		/* The string to insert... */
	    i = strloc(t, '}') ;
	    if (i == 0) {	/* V4.8: accept whole record */
		pf = &Wrecord;
	    }
	    else if ((i == 1) && (t[0] == '#')) {	/* Counter */
		pf = &Crecord;
	    }
	    else if ((i==5) && (strncmp(t, "recno", 5) == 0)) { /* V5.72 */
		pf = &Crecord;
	    }
	    else {
	        junk[0] = t[i]; t[i] = 0 ;
	        pf = theCol(T, t, flags, 1) ;
		if (flags[0]&F_XMAP) /* Sub-element */
		    map->asubs[nsubs++] = strchr(t, '$') - text;
	        t[i] = junk[0];
	    }
	    t += i;
	    if (*t) t++ ;
	    if (pf->ff != 'X') { *(af++) = pf; flags++; continue; }
	}

	else if ((*t == '$') && (t[1] == '#')) {	/* Counter */
	    pf = &Crecord;
	    t += 2;
	}
	else if ((*t == '$') && isdigit(t[1])) {
	    t[0] = '#';
	    for (i=1; isdigit(t[i]); i++) ;
	    junk[0] = t[i]; t[i] = 0;
	    if (atoi(t+1) == 0)	/* V4.8: accept whole record */
		 pf = &Wrecord;
	    else pf = theCol(T, t, junk+1, 1);
	    t[0] = '$'; t[i] = junk[0];
	    if (pf) { t += i; *(af++) = pf; flags++; continue; }
	}

	else { /* Simple text, until next comment / end of line */
	    for (p=t; isspec[(*p)&255] == 0; p++) ;
	    i =  p-t; /*if (i == 0) i=1; */
	    p = t ;		/* The string to add...	  */
	    if (t[i] == '\n') { ++i ; colno=0; }
	    else colno += i;
	    junk[0] = t[i]; t[i] = 0 ;
	    pf = theCol(T, t, flags, F_vMAP) ;
	    t[i] = junk[0]; 
	    t += i ; if(i==0) t++;
	}

	if(pf->ff != 'X') { *(af++) = pf; flags++; continue; }	/* V5.95 */
	/* Complete the 'X' column */
	/*printf("#...memcpy(%p,%p,%d)\n", px, pf, sizeof(COLINFO)); */
	memcpy(px, pf, sizeof(COLINFO)) ;
	px->label = s ;
	if(pf->value) memcpy(s, pf->value, pf->len); 	/* memcpy(s, p, i) ; */
	px->value = (char *)0 ;
	s += pf->len /* i */; 
	*(s++) = 0 ;
	*(af++) = px++ ;
	flags++;
    }

    if (tofree) free(tofree);
    return(map);
}

/*==================================================================
		Print TABMAP Explanations
 *==================================================================*/

int tbm_prnotes(TABMAP *tbm, FILE *fout, int P_opt)
/*++++++++++++++++
.PURPOSE  Print out the Notes
.RETURNS  Number of notes
-----------------*/
  /* IN: tbm = The description 		*/
  /* IN: fout = Opened output file 	*/
  /* IN: P_opt= Printing Option		*/
{
  TABINFO *pt; SECTION *ps;
  int imode, nf=0;

    /* Print out the Notes */
    if ((pt = tbm->tabform)) {
        imode = strloc(modes, tbm->mode) ;
    	ps=pt->section;
    	for (ps=ps->next; ps->class == SEC_Note; ps=ps->next) {
    	    /* if (nf == 0) pr_com(" ", fout, imode) ; */
	    pr_com(ps->header, fout, imode) ;
	    fputc('\n', fout) ;
            if (ps->abody && (ps->header != ps->abody)
		          && (ps->abody < ps->ebody))
	        pr_com(ps->abody, fout, imode), fputc('\n', fout) ;
            nf++;
    	}
    	if (nf) fprintf(fout, "%s%s\n", com0[imode], RULE(tbm->mode)+1) ;
    }
    return(nf);
}

int tbm_prhead(TABMAP *tbm, FILE *fout, int P_opt)
/*++++++++++++++++
.PURPOSE  Print out the "Table Mapping"
.RETURNS  Max length of 1 line
-----------------*/
  /* IN: tbm = The description 		*/
  /* IN: fout = Opened output file 	*/
  /* IN: P_opt= Printing Option		*/
{
  static CATINFO *last_cat ;	/* Catalog edited */
  static BUFFER buf=BUF_Init(char, 512);
  static char *catitle;
  int i, m, nf, maxlen, nb, imode;
  int len=0, lglab, lgunt, lgprf;	/* Length: label, unit, prefix */
  char *p, *com_, outline=0, *Rule;
  COLINFO *pf, **af, **bf;
  TABINFO *pt;
  char fmt_tit[60];
  int contindent=0;	/* Identation for continuation */

    /* pref = tbm->prefix; */
    imode = strloc(modes, tbm->mode) ;
    com_ = com0[imode];	/* e.g. '#' on each line */
    Rule = RULE(tbm->mode);

    /* Write document header */
    if (dochead[imode]) {
	fprintf(fout, dochead[imode], VERSION, date('T'));
	dochead[imode] = (char *)0;
	conclude = doctail[imode];
    }

    /* Write out title of catalog, of file */
    /* Here we set i to 0 if catalog not yet edited. */
    if (theCat != last_cat) {
	last_cat = theCat;
	if (catitle) {
	    for (i=0; isgraph(catitle[i]) && (theCat->fbuf[i] == catitle[i]);
		    i++ );
	    if (isgraph(catitle[i]) || isgraph(theCat->fbuf[i])) i=0;
	}
	else i = 0;
	if (i == 0) { int c=' ';
	    if (catitle) free(catitle);
	    catitle = strdup(theCat->fbuf);
	    /* Squeeze blanks */
	    for (p=catitle; isprint(*p); p++) {
		if ((*p == ' ') && (c == ' ')) continue;
		catitle[i++] = c = *p;
	    }
	    if (catitle[--i] != ' ') ++i;
	    catitle[i] = 0;
	    /*fprintf(fout, "%s%s\n", com_, catitle); */
	    /*fprintf(fout, "%s%s\n", com_, Rule+1); */
	    /* for (i=0; isgraph(catitle[i]); i++) ; catitle[i] = 0;
	    pr_com(catitle, fout, imode);   -- Name of Catalog */
	    BUF_Clear(&buf); buf_text(catitle, &buf, imode);
	    /* Separate code of catalog */
	    for (i=0; isgraph(buf.buf[i]); i++) ; buf.buf[i]=0;
	    fprintf(fout, cat_titles[imode], buf.buf, buf.buf+i+1);
	}
    }
    if (theFile) {
	/* Write the full catalog table */
#if 0
	fprintf(fout, "%s", com_);
	if (catitle) {
	    i = strloc(catitle, ' '); catitle[i]=0;
	    fprintf(fout, "%s/", catitle);
	    catitle[i]=' ';
	}
	fprintf(fout, "%s %s (%d records)\n", theFile->dfile,
		theFile->text, theFile->records);
#else
	BUF_Clear(&buf);
	if (catitle) {
	    i = strloc(catitle, ' '); catitle[i]=0;
	    BUF_AppendString(&buf, catitle);
	    BUF_AppendString(&buf, "/");
	    catitle[i]=' ';
	}
#if 0	/* V5.92 */
	buf_text(theFile->dfile, &buf, imode);  BUF_AppendString(&buf, " ");
#else
#endif
	buf_text(    tbm->dfile, &buf, imode);  BUF_AppendString(&buf, " ");
	buf_text(theFile->text,  &buf, imode);  BUF_AppendString(&buf, " ");
	for (i=0; isgraph(buf.buf[i]); i++) ; buf.buf[i++]=0;
	/* V5.65: Remove the multiline */
	p = buf.buf+i; 
	i = strloc(p, '\n') ;
	while (p[i]) { char *b;
	    for (b=p+i; isspace(*b); b++) ; --b;
	    memmove(p+i, b, strlen(b)+1);
	    p[i] = ' '; 
	    i = strloc(p, '\n');
	}
	fprintf(fout, tab_titles[imode], buf.buf, p, theFile->records);
#endif
    }

    /* fprintf(fout, "%s\n", rule('-')); */
    if ((pt = tbm->tabform)) {
    	if (pt->equinox) fprintf(fout, "%sEquinox=%s%s\n",
	   com_, pt->equinox, com1[imode]);
    	if (pt->epoch)   fprintf(fout, "%sEpoch=%s%s\n",
	   com_, pt->epoch  , com1[imode]);
    }

    /* In Paragraph mode, edit <label> -- but not in tabular mode */
    outline = tbm->mode == 'p';
    /* ... still in paragraph mode, the F_0xMAP flag indicates that
     * only notes are written out. in the header */
    /* --- not accepted in HTML comments */

    /* Count the length of label, unit, prefix */
    lglab = lgunt = lgprf = i = 0 ;
    for (m = tbm->nfmt, af = bf = tbm->afmt; --m >= 0; af++) {
    	pf = *af;
	if ((p = pf->unit)) { len = strlen(p); lgunt = MAX(lgunt, len); }
	p = pf->label ;
        if (*p == '\n') i = 0 ;
	else if (pf->ff == 'X') {
	    len = i == 0 ? strlen(p) : 0 ;
	    lgprf = MAX(lgprf, len) ;
	}
	else { len = strlen(p); lglab = MAX(lglab, len); }
    }
    if (lglab < 9) lglab = 9;
    if (lgunt < 5) lgunt = 5 ;
    sprintf(fmt_tit, "%%s %%%ds %%-6s %%-%ds %%s\n", lglab, lgunt) ;
    contindent = lglab + lgunt + 10;

    nf = 0;		/* Number of Fields */
    maxlen = 0;		/* Size of 1 line   */
    fflush(fout);
    if (*com1[imode]) {	/* Just comment once to open, once to close */
	fprintf(fout, "%s\n", com_) ;
        com_ = "" ;
    }
    if (tbm->flags&F_0xMAP) ;
    else for (m = tbm->nfmt, af = bf = tbm->afmt; --m >= 0; af++) {
    	pf = *af;
    	if (strchr(pf->label, '\n')) {	/* Explain Previous Headers */
	    len = 0 ;
	    /* Preface in PARFILE: don't start by a newline */
	    if (outline && (nf == 0)) continue ;
    	    if (nf > 4) { 		/* Add Titles of Explanation */
		fprintf(fout, "%s%s\n", com_, Rule+1) ;
    	    	fprintf(fout, fmt_tit, /* "%s %9s %-6s %-7s %s\n", */
    	    	    com_, "Label", "Format", "Unit", "Explanations") ;
		fprintf(fout, "%s%s\n", com_, Rule+1) ;
	    }
    	    for (i=0; bf < af; bf++) {
    		pf = *bf;
    	    	if (pf->ff == 'X') { 	/* Compute the position in line */
    	    	    for (p = pf->label; *p; p++) {
    	    	    	if (*p == '\n') i = 0;
    	    	    	else i++;
    	    	    }
    	    	    continue;
	    	}
		if (outline == 0) {	/* All modes except par */
		  char *p, *e;
		    p = pf->text;
		    if ((e = strchr(p, '\n')))	/* Several lines */
			*e = 0;
		    fprintf(fout, fmt_tit, /* "%s %9s %-6s %-7s %s\n", */
			com_, pf->label, fmt_ed(pf), pf->unit, p);
		    while (e) {		/* V4.91: Indent  explanations */
			*(e++) = '\n';
			p = e;
			if ((e = strchr(p, '\n'))) *e = 0;
			while (*p == ' ') p++;
			fprintf(fout, "%s ", com_);
			fwrite(blanks, 1, contindent, fout);
			fprintf(fout, "%s\n", p);
		    }
		    /*
		    fprintf(fout, "%s ", com_) ;
		    for (nb=lglab-strlen(pf->label); --nb>=0;) fputc(' ', fout);
		    fprintf(fout, "%s", pf->label);
		    */
		}
		else fprintf(fout, " %-6s %-7s %s\n",
    	    		fmt_ed(pf),pf->unit, pf->text);
		i += pf->len*pf->rep;
    	    }
    	    if (nf > 4)  		/* Add Titles */
    	    	fprintf(fout, "%s%s\n", com_, Rule+1) ;
	    maxlen = MAX(maxlen, i);
    	    pf = *af; nf = 0;
    	    /* fprintf(fout, "%s", pf->label+1);  */
	    len = 0 ;
    	}
    	else switch(pf->ff) {	/* Not a newline */
    	    case 'X':
		if(outline) fprintf(fout, "%s", pf->label),
		    len += strlen(pf->label);
		break;
    	    default:
		nf++;
		if(outline) {
		    fprintf(fout, "<%s>", pf->label);
		    nb = lglab + lgprf -len - strlen(pf->label);
		    while (--nb>=0) fputc(' ', fout);
		}
		break;
    	}
    }

    /* Print out the Notes */
    if (tbm->tabform && ((P_opt&P_note1)==0)) {
	tbm_prnotes(tbm, fout, P_opt);
	if (maxlen<80) maxlen=80;
    }
    if (*com1[imode]) fprintf(fout, "%s\n", com1[imode]) ;

    return(maxlen);
}

/*==================================================================
		Define a Description
 *==================================================================*/

static int nullcol(char *text, int len, int isnum)
/*++++++++++++++++
.PURPOSE  Verify a field is all blank / null
.RETURNS  1 (true) / 0 (false)
-----------------*/
{
  char *p, *e;
    for (p=text, e=text+len; *p && (p<e); p++) {
        if (*p != ' ') {
	    if (!isnum) return(0) ;	/* Not Null ! */
	    /* A numeric field may be NULL with a least 2 minuses */
	    if (ispunct(*p) && (p[1] == p[0])) ;
	    else return(0) ;
	}
    }
    return(1) ;
}


static int isnullcol(COLINFO *pf, char *text, int len)
/*++++++++++++++++
.PURPOSE  Verify a field is NULL
.RETURNS  1 (true) / 0 (false)
.REMARKS  F_REQD and F_NULL flags set ==> test specified value
-----------------*/
{
  double *dlim, atof();
  int8 *llim;
  char x;
  int stat;
    /*if (pf->elims&F_REQD) return(0);*/
    if ((pf->elims&(F_NULL|F_REQD)) == (F_NULL|F_REQD)) {
	/* Explicit NULL value (mod. V5.71) */
	x = text[len]; text[len] = 0;
	if (pf->elims & F_DBLE) {
	    dlim = (double *)pf->lims;
	    stat = atof(text) == dlim[2];
	}
	else {
	    llim = (int8  *)pf->lims;
	    stat = atoll(text) == llim[2];
	}
	text[len] = x;
    }
    else return(nullcol(text, len, pf->ff != 'A')) ;
    return(stat) ;
}


static int buf_addcol(BUFFER *b, char *text, char *option, COLINFO *pf)
/*++++++++++++++++
.PURPOSE  Append a ${...}
.RETURNS  0
.REMARKS  ${---} column (constant column) replaced by text ${+offset(len)}
-----------------*/
  /* MOD: b = Buffer to use */
  /* IN: text = Text to insert */
  /* IN: option = Additional option */
{
    BUF_AppendString(b, "${") ;
    if (option) BUF_AppendString(b, option) ;
    if ((text[0] == '-') && pf)  {	/* Constant column */
	char buf[24];
	sprintf(buf, "+%d(%d)", pf->b0, 1+(pf->b1-pf->b0));
	BUF_AppendString(b, buf);
    }
    else BUF_AppendString(b, text) ;
    BUF_AppendString(b, "}") ;
    return(0) ;
}

static int tbm_prule(TABMAP *tbm, BUFFER *b, int car)
/*++++++++++++++++
.PURPOSE  Print a Rule from TABMAP
.RETURNS  1 (1 line)
-----------------*/
  /* IN: tbm = The description 	*/
  /* IN: fout = Opened output file */
  /* IN: car = Character to use (0 => leave the | ) */
{
  int m, imode;
  char *p, c;
  COLINFO *pf, **af;

    imode = strloc(modes, tbm->mode) ;
    if (rules[imode]) {
	BUF_AppendString(b, rules[imode]) ;
	return(1) ;
    }
    if (car && (! isprint(car))) car = '-';
    for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++) {
    	pf = *af;
    	if (pf->ff == 'X') {
	    if (tbm->mode == 'c') BUF_AppendString(b, pf->label) ;
    	    else if (tbm->mode == 'a') for (p = pf->label; *p; p++) {
    	    	if ((c = car)) ;
    	    	else c = (*p == ' ' ? '-' : *p);
		if ((*p == '\n')) c = '\n';	/* Respect newline always */
		BUF_AppendChars(b, c, 1) ;
	    }
    	    continue;
    	}
	BUF_AppendChars(b, '-', pf->len*pf->rep) ;
    }
    return(1);
}

static int print_message(TABMAP *tbm, FILE *fout, char *text)
/*++++++++++++++++
.PURPOSE  Print a message (error message)
.RETURNS  0 (no message printed) / 1
.REMARKS  Added V5.64
-----------------*/
{
  char *fmt, *s;
  int imode;

    imode = strloc(modes, tbm->mode);
    fmt = fmsg[imode];
    if (!fmt) return(0);
    s = strstr(fmt, "%s");
    if (!s) { fprintf(fout, fmt, s); return(1); }
    fwrite(fmt, 1, (s-fmt), fout);
    if (aspecs[imode]) { char **aspec=aspecs[imode], *p; int k;
	for (p=text; *p; p++) {
            for (k=0; aspec[k] && (*aspec[k] != *p) ; k++) ;
            if (aspec[k]) fprintf(fout, "%s", aspec[k]+1) ;
	    else fputc(*p, fout);
	}
    }
    else fprintf(fout, "%s", text);
    fprintf(fout, "%s", s+2);
    return(1);
}

#define No_Label(pf) (pf->ff == 'X')||(pf->label[0] == '-')

static int make_title(TABMAP *tbm)
/*++++++++++++++++
.PURPOSE  Create the Title (prefix / postfix)
.RETURNS  The number of title lines
.REMARKS  For a paragraph file, no header...
-----------------*/
  /* IN: tbm = The description 	*/
  /* IN: fout = Opened output file */
{
  BUFFER *b ; char buf[60];
  int i, m, w, ml, nl, ll, lines, nfmt, nfmt2, nb, imode, multicol=0;
  COLINFO *pf, **af, **ef, **qf, **vf;
  int *llines, *pi;
  char hash=0;
  char *p, *e, *o;

    if (tbm->mode == 'p') return(0) ;
    imode = strloc(modes, tbm->mode) ;
    b = BUF_Open(char, 512, 512);
    lines = 0;
    o = head_opt; if (*o == 'H') o++;

    if (tbm->mode == 'c') {	/* CSV: Output the Titles + Units */
	  for (; *o; o++) {
	      if (hash) BUF_AppendItem(b, char, &hash);
	      switch(*o) {
	      case '1': case 'l':	/* Labels (column names) */
	        lines++;
                for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++) {
	            pf = *af;
	            BUF_AppendString(b, pf->label) ;
	        }
	        continue;
	      case 'u':		/* Units */
	        lines++;
	        for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++, pi++) {
	            pf = *af;
	            if (pf->ff == 'X') BUF_AppendString(b, pf->label) ;
	            else if (pf->unit) BUF_AppendString(b, pf->unit ) ;
	        }
	        continue;
	      case '-':		/* Line */
	        lines++;
	        tbm_prule(tbm, b, '-') ;
	        continue;
	      case '#':  hash = *o; continue; 
              case 'f': case 't':	/* Table name */
	        if (theFile) BUF_AppendString(b, theFile->dfile);
	    }
	}
	BUF_AppendString(b, "");	/* Be sure b is zero-terminated */
	tbm->prefix = b->buf ;
	free(b) ;
        return(lines) ;
    }

    if (tbm->mode == 'H') {	/* HTML Titles */
	tbm->postfix = "</TABLE>\n" ;
        BUF_AppendString(b, "<TABLE class='sort'") ;
	if (tbm->bars) {
	    sprintf(buf, " BORDER=%d CELLSPACING=%d CELLPADDING=%d",
		    tbm->bars, 2, 2);
	    BUF_AppendString(b, buf);
	}
        BUF_AppendString(b, ">\n") ;
	lines = 1 ;
	o = strchr(head_opt, 'u');	/* Not NULL to add Unit */
        for (af = tbm->afmt, ef = af + tbm->nfmt; af<ef; af = qf) {
    	    pf = *af; qf = af+1;
	    p = pf->label ;
    	    if (pf->ff == 'X') {
		while (*p) {	/* Actual mapping */
		    i = strloc(p, '<') ;
		    if (i == 0) {
		        i = 1 ;
		        e = p + strloc(p, '>') ;
		        if (p[i] == '/') i++;
		        if ((p[i] == 'T') && (p[i+1] == 'D')) {
		            ++i ;
		            BUF_AppendItems(b, char, p, i);
		            BUF_AppendString(b, "H") ;
		        }
		        else BUF_AppendItems(b, char, p, e-p) ;
		        p = e ;
		        continue ;
		    }
		    BUF_AppendItems(b, char, p, i);
		    p += i;
		}
		continue;
	    }
	    /* Here we have a 'normal' column (column with contents) */
		/* How many columns to merge ? */
	    while (qf<ef) {
		if (((*qf)->ff == 'X') && strchr((*qf)->label, '<'))
		    break;
		qf++;
	    }
	    /* Write labels */
	    nb = 0;  /* Set to number of blanks to issue */
	    for (vf=af; vf<qf; vf++) {
		pf = *vf;
		if (No_Label(pf)) continue;
		if (pf->flags&F_ACCU)
		    continue;
		if (nb) BUF_AppendChars(b, ' ', 1); nb++;
		p = pf->label;
	        if (p[1] == '_') {   /* Write a smaller label */
		    BUF_AppendString(b, "<font size=\"-1\">(");
		    BUF_AppendChars(b, *p, 1);
		    BUF_AppendString(b, ")</font>");
		    continue;
		}
		if (pf->flags&F_ACCU2) { /* Remove suffix */ int c;
		    i = strbloc(p, '.');
		    if (i<0) i=strlen(p)-1;
		    c = p[i]; p[i] = 0;
		    buf_text(p, b, imode);
		    p[i] = c;
		}
		else buf_text(p, b, imode) ;
	    }
	    if (o) {	/* Add the unit, only if label present */
		BUF_AppendString(b, "<BR>"); nb = 0;
		for (vf=af; vf<qf; vf++) {
		    pf = *vf;
		    if (No_Label(pf)) continue;
		    if (pf->label[1] == '_') continue;
		    p = pf->unit;
		    if (!*p) continue;
		    if (pf->flags&(F_ACCU|F_ACCU2)) {
			if ((pf->flags&F_ACCU2) && nb)
			    BUF_AppendString(b, " ");
			if (*p != '-') {  /* Actual unit */
			    /* Convert arcmin arcsec deg */
			    if (p[0] == 'a') p += 3;
			    BUF_AppendItem(b, char, p);
			}
		    }
		    else { /* Standard Unit */
			if (nb) BUF_AppendString(b, " ");
			BUF_AppendString(b, pf->unit);
		    }
		    nb++;
		}
	    }
	}
        BUF_AppendString(b, "");	/* Be sure b is zero-terminated */
	tbm->prefix = b->buf ;
	free(b) ;
        return(lines) ;
    }

    if ((tbm->mode == 'h') || (tbm->mode == 't')) {	/* HTX / latex */
	lines = 1 ;
	if (tbm->mode == 't') o = (char *)0;
	else o = strchr(head_opt, 'u');		/* Not NULL to add Unit */
	if (tbm->mode == 'h') {
	    e = "p" ;		/* Paragraph option */
	    tbm->postfix = "\\End{tabular}\n";
	    BUF_AppendString(b,
	       "\\Beg{TABULAR}{\\bg{white} CELLPADDING=2}") ;
	}
	else {
	    e = "p{4em}" ;
	    tbm->postfix = "\\hline\\end{tabular}\n";
	    BUF_AppendString(b,
	      "\\begin{tabular}\\hline") ;
	}
        BUF_AppendString(b, "{") ;
	if ((w=tbm->bars)) BUF_AppendChars(b, '|', w) ;
	/* Define the Options for Alignement */
        for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++) {
    	    pf = *af;
	    p = pf->label ;
	    if (No_Label(pf)) continue ;
	    if (pf->ff == 'A') p = pf->len > nowrap ? e : "l" ;
	    else p = "r" ;
	    BUF_AppendString(b, p) ;
	    if (tbm->bars) BUF_AppendString(b, "|") ;
	}
	/* Already one | added */
	if (w>1) BUF_AppendChars(b, '|', w-1) ;
	BUF_AppendString(b, "}\n") ;
	if (tbm->mode == 't') BUF_AppendString(b, "\\hline\n") ;
	/* Add the Titles */
        for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++) {
    	    pf = *af;
	    p = pf->label ;
    	    if (pf->ff == 'X') { 	/* Mod. V4.92 */
		if (multicol && ((*p == '&') || strncmp(p, "\\\\", 2) == 0)) {
		    BUF_AppendString(b, /*"{"*/"}");
		    multicol=0;
		}
		BUF_AppendString(b, p);
		p = ""; 		/* Don't add unit... */
	    }
	    else {		/* Don't write label if '-' */
		if (p[1] == '_') {
		    if (af[-1]->ff != 'X') p = "" ;
		    if (af[01]->ff != 'X') p = "" ;
		}
		if ((tbm->mode == 'h') && (multicol == 0)) {
		    /* htx: write the titles in bf */
		    multicol = 1;
		    BUF_AppendString(b, "\\multicolumn{1}{c}{{\\bf " /*"}}"*/);
		}
	        buf_text(p, b, imode) ;
		if (multicol) BUF_AppendString(b, /*"{"*/"}");
	    }
	    if (o && *p) {	/* Add the unit, only if label present */
		BUF_AppendString(b, "\\br ");
		p = pf->unit;
		if (p) BUF_AppendString(b, p);
	    }
	}
	if (multicol) { BUF_AppendString(b, /*"{"*/"}"); multicol = 0; }
	if (tbm->mode == 't') {
	    /* LaTeX: Add a line with the units */
	    if (strchr(head_opt, 'u')) {
	        for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++) {
		    pf = *af;
		    if (pf->ff == 'X')  BUF_AppendString(b, pf->label);
		    else if ((p=pf->unit)) buf_text(p, b, imode);
		}
		/* BUF_AppendString(b, "\\\\\n"); */
	    }
	    BUF_AppendString(b, "\\hline\n") ;
	}

        BUF_AppendString(b, "");	/* Be sure b is zero-terminated */
	tbm->prefix = b->buf ;
	free(b) ;
        return(lines) ;
    }

	/* Titles in Ascii Format: we keep an array with
	   - number of lines for 1 label
	   - length of label
	   - number of written bytes
	*/
    nfmt = tbm->nfmt; nfmt2 = nfmt * 2 ;
    pi = llines = (int *)malloc(nfmt * 3 * sizeof(int));
    ml = i = 0;	/* maxlen = linesize, ml = max. lines */
    for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++, pi++) {
    	pf = *af;
    	if (No_Label(pf))   		/* No label required */
    	    *pi = 0;
    	else {
    	    w = pf->len*pf->rep;	/* Width of Field */
    	    p = pf->label;
	    ll = strlen(p);
	    if (p[1] == '_') {		/* Remove the label unless
					   surrounded by ColSep	*/
		if ((af[-1]->ff == 'X') && (af[1]->ff == 'X') &&
		    isgraph(*(af[-1]->label)) && isgraph(*(af[1]->label))) ;
		else ll = 2;
	    }
	    if (pf->flags&F_SIGN) ll=0 ;	/* Don't edit DE- */
	    pi[nfmt] = ll;
    	    nl = (ll + w -1) / w;
    	    ml = MAX(ml, nl);
    	    *pi = nl;	/* We keep how many lines are required for the label */
	    pi[nfmt2] = nl > 1 ? ll - nl*w : 0 ;	/* Blanks on 1st line*/
    	    /* printf(" %s=%d", p, *pi); */
    	}
    }
    lines = ml + 2;		/* Total lines = 2*Rule + other */
    /* printf("\nTotal lines: %d\n", lines); */

    	/* Write now the "Rule" -- also at end ! */
    tbm_prule(tbm, b, '-');
    BUF_SaveString(b, "") ;
    b->offset = b->used ;	/* Start of prefix */
    tbm_prule(tbm, b, '-');

    	/* Write now the Title */
    while (--ml >= 0) {
    	pi = llines;
    	for (m = tbm->nfmt, af = tbm->afmt; --m >= 0; af++, pi++) {
    	    pf = *af;
    	    if (pf->ff == 'X') {
		BUF_AppendString(b, pf->label);
    	    	continue;
    	    }
    	    w  = pf->len*pf->rep;	/* Width of Field */
    	    i  = *pi - ml;		/* 1 = First line , etc  */
	    nb = pi[nfmt2] ;		/* Current offset in lab */
    	    if (i > 0) {		/* We've to write a part */
    	    	p  = pf->label;
    	    	ll = pi[nfmt];		/* Length of label */
		while (nb<0) BUF_AppendChars(b, ' ', 1), w--, nb++ ;
    	    	p += nb;
		/* V4.81: Right alignement if a non-alphabetic field */
		if ((ml==0) && (pf->ff!='A') && (pf->rep<2) && (ll<w)) {
		    BUF_AppendChars(b, ' ', w-ll);
		    w = ll;
		}
    	    	while ((w>0) && (nb<ll))
    	    	    BUF_AppendChars(b, *p, 1), p++, w--, nb++;
	    }
	    pi[nfmt2] = nb ;
	    /* Don't output the trailing blanks when last field */
	    if (m) {
	        if (*(af[1]->label) != '\n')
		    BUF_AppendChars(b, ' ', w) ;
	    }
    	}
    }

    	/* Another (transparent) rule */
    tbm_prule(tbm, b, 0);
    free(llines);

    /* Report the strings in the TABMAP structure */
    BUF_AppendString(b, "");	/* Be sure b is zero-terminated */
    tbm->postfix = b->buf ;
    tbm->prefix  = b->buf + b->offset;
    free(b) ;
    return(lines);
}

int tbm_pr1(TABMAP *tbm, FILE *fout, char *buf)
/*++++++++++++++++
.PURPOSE  Print a Record according to TABMAP
.RETURNS  1
-----------------*/
  /* IN: tbm = The description 	*/
  /* IN: buf = The record to print */
  /* IN: fout = Opened output file */
{
  int i, k, lb, len, stat, imode; 	/* stat: 0 / 1=not null / 2=null */
  char *p, *e, *m, **aspec, x;
  COLINFO *pf, **af;
  char *lab ;

    len = strline(buf);
    x = buf[len]; buf[len] = 0;
    lab = (char *)0 ;
    imode = strloc(modes, tbm->mode) ;
    aspec = aspecs[imode] ;
    stat = 0 ;
    for (af=tbm->afmt, i=tbm->nfmt, m=tbm->aflags; --i >= 0; af++, m++) {
	pf = *af;
	if (pf->ff == 'W') {	/* Whole record */
	    if (stat == 0) {
		if (lab) fprintf(fout, "%s", lab);
	        lab = pf->label;
	    }
	    else {
		if (stat&1) {
		    if (lab) fprintf(fout, "%s", lab) ;
		}
		lab = (char *)0 ;
		if (strchr(pf->label, '\n')) stat = 0 ;
		fwrite(buf, 1, len, fout);
		/* Fill with blanks */
		for (lb=pf->len - len; --lb>=0; ) fputc(' ', fout);
	    }
	    continue;
	}
	if (pf->ff == 'X') {	/* Text */
	    if (stat == 0) {
	        if (lab) fprintf(fout, "%s", lab);
		lab = pf->label ;
		if (TRmod && (strncmp(lab, TRstyle, iTRmod) == 0))
		    lab[iTRmod] = ('0'|TRmod) - (recno%TRmod);
	    }
	    else {
	        if (stat&1) {
		    if (lab) fprintf(fout, "%s", lab) ;
		    fprintf(fout, "%s", pf->label) ;
		}
		lab = (char *)0 ;
		if (strchr(pf->label, '\n')) stat = 0 ;
	    }
	    continue ;
	}
	lb = 1 + (pf->b1 - pf->b0);	/* Width of Field */
	/* TODO: EXTRACT subfield ---------- */
	/* Ignore a field which is empty */
	if (tbm->flags&F_0MAP) {
	    if (len < pf->b0) stat |= 2 ;	/* null */
	    else if (nullcol(buf + pf->b0, lb, pf->ff != 'A')) stat |= 2 ;
	    else stat |= 1 ;
	}
	else stat = 1 ;
	if (stat&1) {
	    if (lab) fprintf(fout, "%s", lab), lab = (char *)0;
	    p = buf + pf->b0, e = p + lb ;
	    if ((e-buf) > len) e = buf+len ;
	    if (e<p) e = p;
	    if (pf == &(Crecord)) {
		sprintf(arec, "%6d", recno);
		p = arec, e = arec+strlen(arec);
	    }
	    if (null_value && isnullcol(pf, p, e-p))	/* V4.4 */
		p = null_value, e = p+strlen(p);
	    if (bopt || ((*m)&F_tMAP))
		while (e>p) { if (*--e != ' ') { ++e; break; } }
	    if (bopt || ((*m)&F_hMAP))
		while (p<e) { if (*p != ' ') break; p++; }
	    while (p<e) {
		if (aspec) {
		    for (k=0; aspec[k] && (*aspec[k] != *p) ; k++) ;
		    if (aspec[k]) fprintf(fout, "%s", aspec[k]+1) ;
		    else fputc(*p, fout);
		}
	        else fputc(*p, fout);
		p++; lb--;	/* Corrected V4.3 */
	    }
	    /* lb -= (e-(buf+pf->b0)) ;	-- Remaining blanks */
	    if ((lb>0) && i && (!bopt) && (!((*m)&F_tMAP))) {
	        /* Add blanks; not required for last field */
	        if (*(af[1]->label) != '\n')
	    	    while (--lb>=0) fputc(' ', fout);
	    }
	}
    }
    buf[len] = x ;
    return(1);
}

/*==================================================================
		Default Mapping of Labels into Text
 *==================================================================*/

static int cplx_label(COLINFO *col, int keep_cols)
/*++++++++++++++++
.PURPOSE  Check whether label related to a complex number
.RETURNS  1 (RA/LON) 2 (DE/LAT) / 4 (non-position, but complex)
-----------------*/
{
  char *label;	/* IN: Label of the column */
    label = col->label;
    if (!keep_cols) {
        if (strmatch("RA*", label)) return(1);
        if (strmatch("DE*", label)) return(2);
        if (strmatch("[GES][Ll][Oo][Nn]", label))  return(1);
        if (strmatch("[GES][Ll][Aa][Tt]", label))  return(2);
    }
    if (col->flags&(F_ACCU|F_ACCU2|F_SSEP)) return(4/*1*/) ;
    return(0);
}

static COLINFO *cplx_next(TABINFO *tab, COLINFO *col, int prefix)
/*++++++++++++++++
.PURPOSE  Find the next column in a complex set
.RETURNS  Next col / NULL
-----------------*/
{
  char *label;	/* IN: Label of the column */
  int len, i;
  COLINFO *col1;
    /* Prefix: 'l' must preced, 'u' must follow */
    label = col->label;
    if (prefix) {
	i = (col - (tab->afmt));
	if ((prefix == 'l') && (i>0))
	     col1 = col-1;
	else if (++i<tab->nfmt) col1 = col+1;
	else return((COLINFO *)0);
	if ((col1->label[0] == prefix) && (col1->label[1] == '_')
		&& (strcmp(label, col1->label+2) == 0))
	    return(col1);
	return((COLINFO *)0);
    }
    if (col->flags&(F_ACCU|F_ACCU2|F_SSEP)) ;
    else return((COLINFO *)0);
    /* Fond out length for label comparison */
    label = col->label;
    len = strlen(label);
    for (i=len-1; (i>0) && (label[i] != '.'); i--) ;
    if (i>0) len += 1;
    else {     /* Accept 1letter or 2letter prefix */
        len -= 1;
        if (strchr("dcm", label[len])) {
    	    --len;
    	    if (strchr("dcm", label[len])); /* Deci centi milli */
    	    else ++len;
        }
    }
    i = col - (tab->afmt);
    i = tab->nfmt - i;
    if (i>0) for (col1=col+1; --i>=0; col1++) {
	if (col1->label[0] == '-') continue;	/* --- ignore such label */
        if (col1->flags&(F_ACCU|F_ACCU2|F_SSEP));
        else return((COLINFO *)0);
        if (strncmp(col1->label, label, len) == 0)
    	    return(col1);
    }
    return((COLINFO *)0);
}

static COLINFO *col_choice(TABINFO *T, char *list)
/*++++++++++++++++
.PURPOSE  Create the list of columns corresponding to a list (blank-separated)
.RETURNS  An allocated COLINFO / NULL if error
-----------------*/
{
  COLINFO *new, *col, *cc, *last_col;
  int ncol = 0;
  int nerr = 0;
  char *v, *p, c, buf[8];

    /* Remove the , in list */
    v = list;
    while((v = strchr(v, ','))) {
	if (v[1] == ' ') { *v = ' '; continue; }
	/* A comma embedded in a word ? Some column names include a comma... */
	for (p=v; (p>list) ; p--) { --p; if(isspace(*p)) { p++; break; }}
	/* Try if the word is known without the comma... */
	*v = 0;
	if (tbf_findlabel(T, p)) *v = ' '; else *v = ',';
	v++;
    }

    /* Interpret the list of columns -- the list is now blank-separated
       Mark the columns used with F_USE2 (set to zero with -b or -k)
    */
    last_col = T->afmt + T->nfmt -1; 	/* V5.66 */
    for (col=T->afmt; col<=last_col; col++)
	col->flags &= ~(F_USER|F_USE2);

    for (p=list; *p; p=v) {
	if (isspace(*p)) { for(v=p; isspace(*v); v++) ; continue; }
	for (v=p; isgraph(*v); v++) ;
	if (*p == '*') { 	/* All columns */
	    continue;
	}
	c = *v; *v = 0;
	col = tbf_findlabel(T, p);
	if ((!col)&&(strlen(p)<sizeof(buf)-2)) {  /* Try e.g. $RA */
	    if ((p[0] == '#') && (p[1] == 0)) {   /* Counter */
		col = &Crecord; 
		col->flags |= F_USER;
	    }
	    else {
	        buf[0] = '$'; strcpy(buf+1, p);
	        col = tbf_findlabel(T, buf);
	    }
	}
	if (!col) { char *a;    /* Could be a comma */
	    if ((a = strchr(p, ','))) {
		*v = c; *a = ' ';
		v = p;
		continue;
	    }
	    fprintf(stderr, "****Column not found: %s\n", p);
	    nerr++;
	    *v = c;
	    continue;
	}
	if (col->flags&F_USER)     /* Column repeated */
	    ncol++;
	col->flags |= F_USER;
	/* Verify preceding/following columns with l_ and u_ */
	if (col->flags&(F_ACCU|F_ACCU2)) {
	    cc = col;
	    while ((cc = cplx_next(T, cc, 0)))
		cc->flags |= F_USE2;
	}
	else if (keep_cols) ;	/* Keep exact columns: don't add limits, u */
	else if (col == &Crecord) ;	/* Counter column */
	else {
	    if ((cc = cplx_next(T, col, 'l')))
		cc->flags |= F_USE2;
	    if ((cc = cplx_next(T, col, 'u')))
		cc->flags |= F_USE2;
	}
	*v = c;
    }
    if (nerr) {
	for (col=T->afmt; col<=last_col; col++)
	    col->flags &= ~(F_USER|F_USE2);
	return((COLINFO *)0);
    }

    ncol += T->nfmt +1;
    new = cc = (COLINFO *)malloc(ncol*sizeof(COLINFO));

    /* Write the columns marked */
    for (p=list; *p; p=v) {
	if (isspace(*p)) { for(v=p; isspace(*v); v++) ; continue; }
	for (v=p; isgraph(*v); v++) ;
	if (*p == '*') { 	/* All columns */
	    for (col=T->afmt; col<=last_col; col++) {
		if (col->flags&(F_USER|F_USE2)) continue;
		*(cc++) = *col;
	    }
	    continue;
	}
	c = *v; *v = 0;
	col = tbf_findlabel(T, p);
	if (!col) {  /* Try e.g. $RA */
	    if ((p[0] == '#') && (p[1] == 0)) { /* Counter */
		col = &Crecord; 
		*(cc++) = *col;
		continue;
	    }
	    buf[0] = '$'; strcpy(buf+1, p);
	    col = tbf_findlabel(T, buf);
	}
	*v = c;
	/* Verify preceding/following columns with l_ */
	if (col!=T->afmt) {
	    --col;
	    if ((col->flags&(F_USER|F_USE2)) == F_USE2) {
		*(cc++) = *col;
		col->flags |= F_USER;
	    }
	    ++col;
	}
	*(cc++) = *col;
	col->flags |= F_USER;
	if (col->flags&(F_ACCU|F_ACCU2)) {
	    while ((col = cplx_next(T, col, 0))) {
		*(cc++) = *col;
		col->flags |= F_USER;
	    }
	}
	else if (col<last_col) {
	    col++;
	    if ((col->flags&(F_USER|F_USE2)) == F_USE2) {
		*(cc++) = *col;
		col->flags |= F_USER;
	    }
	}
    }
    bzero(cc, sizeof(COLINFO));

    /* Reset flags */
    for (col=T->afmt; col<=last_col; col++)
	col->flags &= ~(F_USER|F_USE2);

    return(new);
}

static char *lab_adjust(char *label, int labw)
/*++++++++++++++++
.PURPOSE  Adjust a label according to labw
.RETURNS  Adjusted label
-----------------*/
  /* IN: label = Label of the column */
  /* IN: labw = Width, negative for left-adjustment */
{
  static char buf[80];
  int len, w; char *p;
    w = ABS(labw);
    len = strlen(label);
    if (len >= w) return(label);
    if (len >= (sizeof(buf)-1)) return(label);
    p = buf;
    if (labw > 0) /* Right-Adjustment */
	while (len < w) *(p++) = ' ', len++;
    strcpy(p, label), p += strlen(p);
    while (len < w) *(p++) = ' ', len++;
    return(buf);
}

static char *par_echo(char *label)
/*++++++++++++++++
.PURPOSE  Define the default label mapping
.RETURNS  Its value
-----------------*/
{
  static char skip_prefix[] = "lunf"; /* Labels skipped */
    if (strmatch("RAh", label)) return("RA+DE");
    if (keep_cols) return(label);
    if (label[1] == '_') {
	if(strchr(skip_prefix, *label)) 
	    return(label+2);
    }
    return(label);
}

static int get_w4(char *text, COLINFO *colist, COLINFO *eolist, int *parw)
/*++++++++++++++++
.PURPOSE  Interpret a list of width (a,b...)
.RETURNS  Number of bytes interpreted
-----------------*/
{

  COLINFO *pf;
  char *p; int i, w, iw=0;

    p = text;
    if (*p == '(') p++; 	/* Widths specified */
    for (++p; *p == ' '; p++) ;
    while ((iw<4) && *p) {
	while (*p == ' ') p++;
	i = iw;
	/* TODO eventually: (l=wdith, v=width, u=width */
	parw[i] = atoi(p);
	if (parw[i] == 0) {
	    if (i==3) parw[i] = 1;   /* Explanation to add */
	    else for (pf=colist; pf<eolist; pf++) {
	        if (i==0) w = strlen(pf->label);
	        else if (i==2) w = strlen(pf->unit);
	        else w = pf->len*pf->rep;
	        if (w>parw[i]) parw[i] = w;
	    }
	    if (*p == '-') parw[i] = -parw[i];
        }
        if ((*p == '-')||(*p == '+')) p++;
        while (isdigit(*p)) p++;
        while (*p == ' ') p++;
        if (*p != ',') break;
        while (*p == ' ') p++;
        p++; iw++;
    }
    if (parw[2]>0) parw[2] = -parw[2];   /* unit always left-aligned */

    /* Skip the closing parenthese */
    if ((*text == '(') && (*p == ')')) p++;
    return(p-text);
}

static char *par_sim(char *label)
/*++++++++++++++++
.PURPOSE  Define the SIMBAD-like label mapping
.RETURNS  Its value
-----------------*/
{
  static char buf[80];
  static char J[] = "J", Q[] = "Q", G[] = "G";
  char *p, *t;

    if (strmatch("RA*", label))
    	return(strmatch("*deg", label) ? J : Q);
    if (strmatch("DE*", label))
    	return(strmatch("*deg", label) ? J : Q);
    if (strmatch("[A-Za-z]mag", label)) {
    	sprintf(buf, "M.%c", *label);
    	return(buf);
    }
    if (strdiff("GLON", label) == 0)  { return(G); }
    if (strdiff("GLAT", label) == 0)  { return(G); }
    if (strmatch("Name", label))  return("I");
    if (strdiff("SpType", label) == 0)  return("S");

    for (p = ids; *p; p += 1+strloc(p, ',')) {
    	for (t=label; *p == *t; p++, t++) ;
    	if ((*p == ',') && (*t == '\0')) {
    	    sprintf(buf, "I %s ", label);
    	    return(buf);
    	}
    }
    return(label);
}

/*==================================================================*
 	The choice of the prefixes from standard table labels
	is defined in the array of par_ functions
 *==================================================================*/

typedef struct {
    PCHAR (*fct)();
    char flag[4];
} PCHAR_opt;

static PCHAR_opt par_fct[] = {
    { par_echo,	""  },	/* Default */
    { par_sim, 	"S" },	/* For SIMBAD-Like */
} ;

/*==================================================================*/

static char *tbm_par(TABINFO *T, char *option, COLINFO *colist)
/*++++++++++++++++
.PURPOSE  Define the Defaut Mapping Format in PARAGRAPH form
.RETURNS  Generated code to be compiled
.REMARKS  colist = list of column tp choose (default = all)
	option includes the following:
	* = strip blanks in data
	. = explanations added in output
	0 = don't edit NULL fields
        ^ = indicates the introduction character (% is default)
  (w,v,u) = specifies the width of columns (label, value, unit)
-----------------*/
  /* IN: T = The concerned "Byte-by-Byte" */
  /* IN: option = option includes 'S' for Simbad-like Definitions */
{
  static char blank[] = " ";
  BUFFER *b; COLINFO *pf, *eolist;
  char *p, *pr, *l, *lr, *d, *o;	/* p=preceding label, l=current */
  int i, w, w2; char lc, pc;		/* lc, pc indicate in position  */
  PCHAR (*fct)();
  char strip[4] = "" ;			/* Default: head + tail strip */
  char *sep;
  int parw[4], keep_params;		/* Width of label/value/unit  */

    /* Get the end of list of columns to edit */
    if (colist) { for (eolist=colist; eolist->label; eolist++) ; }
    else        { colist = T->afmt;   eolist = colist + T->nfmt; }
    com0p[0] = '%';

    /* Prepare the buffer to collect the generated instructions */
    b = BUF_Open(char, 512, 512);
    BUF_AppendString(b, "#!form=p");		/* No Rule */

    /* Find out, according to options, how to map names */
    if (! option) option = "";
    memset(parw, 0, sizeof(parw));
    fct = par_echo;				/* Default */
    for (p=option; *p; ) {
	if (*p == '*') {			/* Strip Blanks */
	    for (i=0; *p == '*'; p++) i++ ;
	    if (i > 3) i = 3 ;
	    strip[i] = 0 ;
	    while (--i >= 0) strip[i] = '*' ;
	    continue ;
	}
	if (*p == '^') {	/* Defines the leading char */
	    com0p[0] = *++p;
	    if (*p) p++;
	    continue;
	}
	if (*p == '0') {	/* IGNORE NULL Fields*/
	    p++ ;
	    continue ;
	}
	if (*p == '.') { 	/* Full explanations */
	    get_w4("(+,+,+,+)", colist, eolist, parw);
	    p++;
	    continue;
	}
	if (*p == '(') { 	/* Widths specified */
	    p += get_w4(p, colist, eolist, parw);
	    continue;
	}
    	for (i = 0; i < ITEMS(par_fct); i++) {
    	    if (par_fct[i].flag[0] != *option) continue;
	    fct = par_fct[i].fct;
	    break;
	}
	if (i >= ITEMS(par_fct)) fprintf(stderr,
    	    "#+++tbm_par: option '%c' unknown, taking defaults\n", *p);
	p++;
    }
    if (parw[3]) /* Explanations included */
	BUF_AppendString(b, ".");
    BUF_AppendString(b, option);
    BUF_AppendString(b, "\n");			/* Start empty line */

    /* Strip by default the values */
    if ((!strip[0]) && (!parw[1]))  /* No '*' specified, no value width */
	strip[0] = '*';

    /* Scan the labels, save the result in an expandable buffer */
    p = o = d = "";  lc = pc = 0;
    keep_params = keep_cols || parw[1];
    for (pf = colist; pf < eolist; o=d, p=pf->label, pf++, pc=lc) {
    	pr = p; lr = l = pf->label;
    	if (*p && (p[1] == '_')) pr += 2;
    	if (*l && (l[1] == '_')) lr += 2;
	lc = cplx_label(pf, keep_params); /* 1 for RA part, 2 for DE   */
	if (keep_params);              /* Aligned edition, can't merge */
	else if (strdiff(lr, pr) == 0){	  /* Same root: insert a blank */
    	    				  /* No separation for u_ l_ */
    	    if ((p != pr) && (*p == 'l')) ;
    	    else if ((l != lr) && (*l == 'u')) ;
    	    else BUF_AppendString(b, blank);
	    buf_addcol(b, l, strip, pf) ;
    	    continue;
    	}

    	/* Define how to separate from precedent:
    	   coordinates must stay on same line !
    	   (but always DE after RA, i.e. lc>=pc)
	*/
	sep = "\n";
	if (pc && (lc&3) && (lc >= pc)) { 	/* I have a Coord. */ char *n;
	    /* Replace name Glon Glat -> Gpos, 'RA' by 'Eq' */
	    if ((n = strrchr(b->buf, '\n'))) {
	        if (*++n == com0p[0]) n++;
		if ((n[0]=='R') && (n[1]=='A'))
		    n[0]='E', n[1]='q';
		else if (isgraph(n[1])&&isgraph(n[2])&&isgraph(n[3]))
		    n[1]='p', n[2]='o', n[3]='s';
	    }
    	    sep = blank;
	    if (pf->flags&F_SIGN) sep = "";	/* Sign WITH value */
	}
	BUF_AppendString(b, sep);
	if (*sep == '\n') {
    	    d = (*fct)(l); 			/* Label to write    */
	    if (parw[0]) d = lab_adjust(d, parw[0]);
	    BUF_AppendString(b, com0p);
	    BUF_AppendString(b, d);
	    BUF_AppendString(b, " ");
	}
	w = pf->len*pf->rep;
	if ((w2 = parw[1])) {       /* Need to align the data contents */
	    if ((w2>0) && ((pf->ff == 'A')||(pf->rep>1)))
		w2 = -w2;                    /* Align the text to left */
	}
	if (w2>w) BUF_AppendChars(b, ' ', w2-w);
	buf_addcol(b, l, strip, pf);
	w2 += w;
	if (w2<0)                    /* Left-aligned, add blanks */
	    BUF_AppendChars(b, ' ', -w2);
	if ((w2 = parw[2])) {        /* Edit the Unit */
	    BUF_AppendChars(b, ' ', 1);
	    BUF_AppendString(b, pf->unit);
	    w = strlen(pf->unit) ;
	    w2 += w;
	    if (w2<0) BUF_AppendChars(b, ' ', -w2);
	}
	if (parw[3]) { char *n;       /* Add explanation */
	    BUF_AppendString(b, " # ");
	    n = b->buf+b->used;
	    BUF_AppendString(b, pf->text);
	    /* Remove the newlines in explanations */
	    while ((n = strchr(n, '\n'))) {
		if (!isspace(n[1])) {    /* Leave rules */
		    if (com0p[0]) n[1] = com0p[0];
		    break;
		}
		*(n++) = ' ';
		/* Remove the blanks in explanations */
		for (i=0; isspace(n[i]); i++) ;
		while (n[i]) { n[0] = n[i]; n++; }
		b->used = (n-b->buf);
	    }
	}
    }
    BUF_AppendString(b, "\n");		/* Terminate Paragraph	*/
    p = b->buf;	/* To Return */
    free(b);
    return(p);
}

static char *tbm_ascii(TABINFO *T, char *option, COLINFO *colist)
/*++++++++++++++++
.PURPOSE  Define the Defaut Mapping Format in Ascii Form
.RETURNS  Generated code to be compiled
.REMARKS  Options:  | => column separator
                   || => column separator also beginning / end
-----------------*/
  /* IN: T = The concerned "Byte-by-Byte" */
  /* IN: option = Options TBD */
{
  BUFFER *b; COLINFO *pf;
  char *p, *pr, *l, *lr;	/* p=preceding label, l=current */
  int i, nosep; char lc, pc;	/* lc, pc indicate in position  */
  char *sep, *lastop;
  static char blank[] = " ";
  static char strip[4] = "" ;	/* Default: no strip */

    /* Tell mode */
    strip[0] = 0 ;
    b = BUF_Open(char, 512, 512);
    BUF_AppendString(b, "#!form=a");		/* Tabular Mode */
    BUF_AppendString(b, option);
    BUF_AppendString(b, "\n");
    sep = (char *)0 ;			/* Nothing at line beginning    */
    lastop = "**" ; 			/* Strip last column	*/
    for (p=option; *p; ) {
	if (*p == '*') {			/* Strip Blanks */
	    for (i=0; *p == '*'; p++) i++ ;
	    if (i > 3) i = 3 ;
	    strip[i] = 0 ;
	    while (--i >= 0) strip[i] = '*' ;
	    continue ;
	}
        if (ispunct(*p)) {
            sepchar[0] = *p;
	    while (p[1] == p[0]) p++, sep = sepchar, lastop = (char *)0;
	    p++;
	    continue ;
	}
	fprintf(stderr, "****Unkown option '%c' in '%s'\n", *p, option);
	return((char *)0);
    }
    p = "";  pc = lc = nosep = 0;

    if (colist) { for (pf=colist; pf->label; pf++) ; i = pf-colist; }
    else        { colist = T->afmt; i = T->nfmt; }
    for (pf = colist; --i>=0; pc=lc, pf++, sep++) {
	if (pf->label[0] == '-') {	/* 'Fill' column '---' */
	    buf_addcol(b, pf->label, 0, pf);
	    nosep = 1;
	    continue;
	}
    	pr = p; lr = l = pf->label;
    	if (*p && (p[1] == '_')) pr += 2;
    	if (*l && (l[1] == '_')) lr += 2;
	lc = cplx_label(pf, keep_cols);	/* 1 for RA part, 2 for DE   */
	if (keep_cols) ;
	else if (strdiff(lr, pr) == 0){	/* Same prefix: insert a blank */
	    if (nosep) nosep = 0;
	    else if (sep) {
	        sep = blank;		/* No separation for u_ l_ */
    	        if ((p != pr) && (*p == 'l')) sep = "";
    	        if ((l != lr) && (*l == 'u')) sep = "";
		if (*sep) BUF_AppendString(b, sep);
	    }
	    buf_addcol(b, l, i ? strip : lastop, pf);
	    p = pf->label;
    	    continue;
    	}
	if (nosep) { nosep = 0; sep = ""; }	/* V5.67: col. --- */
	else if (sep && pc && (lc&3) && (lc >= pc)) { /* I have a Coord. */
    	    sep = blank;			/* Just a blank	*/
	    if (pf[-1].flags&F_SIGN) sep = "";	/* Sign WITH value */
	}
	else if (sep) sep = sepchar;
	if (sep && *sep) BUF_AppendString(b, sep);   /* Not in very first */
	buf_addcol(b, l, strip, pf);
    }
    if (!lastop) BUF_AppendString(b, sepchar) ;
    BUF_AppendString(b, "\n");
    p = b->buf;	/* To Return */
    free(b);
    return(p);
}

static char *tbm_csv(TABINFO *T, char *option, COLINFO *colist)
/*++++++++++++++++
.PURPOSE  Define the Defaut Mapping Format in CSV Format
.RETURNS  Generated code to be compiled
.REMARKS  Options:  | => column separator
                   || => column separator also beginning / end
-----------------*/
  /* IN: T = The concerned "Byte-by-Byte" */
  /* IN: option = Options TBD */
{
  char *b, *p, saved ;

    /* Default separator is TAB */
    saved = sepchar[0]; sepchar[0] = '\t';
    b = tbm_ascii(T, option, colist) ;
    sepchar[0] = saved ;
    if ((p = strstr(b, "form=a"))) {
        p = strchr(p, 'a') ;
	*p = 'c' ;
    }
    return(b) ;
}

static char *tbm_latex(TABINFO *T, char *option, COLINFO *colist)
/*++++++++++++++++
.PURPOSE  Define the Defaut Mapping Format in LaTeX Form
.RETURNS  Generated code to be compiled
.REMARKS  Options:  | => add lines
                   || => column separator also beginning / end
-----------------*/
  /* IN: T = The concerned "Byte-by-Byte" */
  /* IN: option = Options TBD */
{
  BUFFER *b; COLINFO *pf;
  char *p, *pr, *l, *lr;	/* p=preceding label, l=current */
  int i; char lc, pc;		/* lc, pc indicate in position  */
  char *sep, *lastop;
  static char blank[] = " ";
  static char strip[4] = "" ;	/* Default: no strip */

    /* Tell mode */
    strip[0] = 0 ;
    b = BUF_Open(char, 512, 512);
    BUF_AppendString(b, "#!form=t");		/* TeX Mode */
    BUF_AppendString(b, option);
    BUF_AppendString(b, "\n");
    sep = (char *)0 ;			/* Nothing at line beginning    */
    lastop = "**" ; 			/* Strip last column	*/
    for (p=option; *p; ) {
	if (*p == '*') {			/* Strip Blanks */
	    for (i=0; *p == '*'; p++) i++ ;
	    if (i > 3) i = 3 ;
	    strip[i] = 0 ;
	    while (--i >= 0) strip[i] = '*' ;
	    continue ;
	}
        if (*p == '|') { while (*p == '|') p++ ; continue; }

	fprintf(stderr, "****Unkown option '%c' in '%s'\n", *p, option);
	return((char *)0);
    }
    p = "";  pc = lc = 0;

    if (colist) { for (pf=colist; pf->label; pf++) ; i = pf-colist; }
    else        { colist = T->afmt; i = T->nfmt; }
    for (pf=colist; --i>=0; pc=lc, p=pf->label, pf++, sep++) {
    	pr = p; lr = l = pf->label;
    	if (*p && (p[1] == '_')) pr += 2;
    	if (*l && (l[1] == '_')) lr += 2;
	lc = cplx_label(pf, keep_cols);	/* 1 for RA part, 2 for DE   */
	if (keep_cols) ;
	else if (strdiff(lr, pr) == 0){	/* Same prefix: insert a blank */
    	    if (sep) {
	        sep = blank;		/* No separation for u_ l_ */
    	        if ((p != pr) && (*p == 'l')) sep = "";
    	        if ((l != lr) && (*l == 'u')) sep = "";
    	        if (*sep) BUF_AppendString(b, sep);
	    }
	    buf_addcol(b, l, i ? strip : lastop, pf);
    	    continue;
    	}
    	if (sep && pc && (lc&3) && (lc >= pc)) { 	/* I have a Coord. */
    	    sep = blank;			/* Just a blank	*/
	    if (pf->flags&F_SIGN) sep = "";	/* Sign WITH value */
	}
	else if (sep) sep = "&";
	if (sep) BUF_AppendString(b, sep);	/* Not in very first */
	buf_addcol(b, l, strip, pf);
    }
    BUF_AppendString(b, "\\\\\n");
    p = b->buf;	/* To Return */
    free(b);
    return(p);
}


static char *tbm_htx(TABINFO *T, char *option, COLINFO *colist)
/*++++++++++++++++
.PURPOSE  Define the Defaut Mapping Format in HTX Format
.RETURNS  Generated code to be compiled
.REMARKS  Options:  | => column separator
                   || => column separator also beginning / end
-----------------*/
  /* IN: T = The concerned "Byte-by-Byte" */
  /* IN: option = Options TBD */
{
  char *b, *p ;
    b = tbm_latex(T, option, colist) ;
    if ((p = strstr(b, "form=t"))) {
        p = strchr(p, 't') ;
	*p = 'h' ;
    }
    return(b) ;
}

static char *tbm_html(TABINFO *T, char *option, COLINFO *colist)
/*++++++++++++++++
.PURPOSE  Define the Defaut Mapping Format in HTML Format
.RETURNS  Generated code to be compiled
.REMARKS  Options:  | => column separator
                   || => column separator also beginning / end
-----------------*/
  /* IN: T = The concerned "Byte-by-Byte" */
  /* IN: option = Options TBD */
{
  BUFFER *b; COLINFO *pf;
  char *p, *pr, *l, *lr;	/* p=preceding label, l=current */
  int i; char lc, pc;		/* lc, pc indicate in position  */
  char *sep, *tag, *lastop;
  static char blank[] = " ";
  static char strip[4] = "" ;	/* Default: no strip */

    /* Tell mode */
    strip[0] = 0 ;
    b = BUF_Open(char, 512, 512);
    BUF_AppendString(b, "#!form=H");		/* HTML Mode */
    BUF_AppendString(b, option);
    BUF_AppendString(b, "\n");
    lastop = "**" ; 			/* Strip last column	*/
    for (p=option; *p; ) {
	if (*p == '*') {			/* Strip Blanks */
	    for (i=0; *p == '*'; p++) i++ ;
	    if (i > 3) i = 3 ;
	    strip[i] = 0 ;
	    while (--i >= 0) strip[i] = '*' ;
	    continue ;
	}
        if (*p == '|') { while (*p == '|') p++ ; continue; }

	fprintf(stderr, "****Unkown option '%c' in '%s'\n", *p, option);
	return((char *)0);
    }
    p = "";  pc = lc = 0;

    BUF_AppendString(b, TRstyle); iTRmod = strloc(TRstyle, '2'); TRmod = 2;

    sep = (char *)0 ;			/* No <TD> yet */
    tag = "<TD>" ;
    if (colist) { for (pf=colist; pf->label; pf++) ; i = pf-colist; }
    else        { colist = T->afmt; i = T->nfmt; }
    for (pf=colist; --i>=0; pc=lc, p=pf->label, pf++, sep++) {
    	pr = p; lr = l = pf->label;
    	if (*p && (p[1] == '_')) pr += 2;
    	if (*l && (l[1] == '_')) lr += 2;
	lc = cplx_label(pf, keep_cols);	/* 1 for RA part, 2 for DE   */
	if (keep_cols) ;
    	if (strdiff(lr, pr) == 0) {	/* Same prefix: insert a blank */
    	    if (sep) {
	        sep = blank;		/* No separation for u_ l_ */
    	        if ((p != pr) && (*p == 'l')) sep = "";
    	        if ((l != lr) && (*l == 'u')) sep = "";
    	        if (*sep) BUF_AppendString(b, sep);
	    }
	    else BUF_AppendString(b, tag) ;
	    buf_addcol(b, l, i ? strip : lastop, pf);
    	    continue;
    	}
    	if (sep && pc && (lc&3) && (lc >= pc)) { 	/* I have a Coord. */
    	    sep = blank;			/* Just a blank	*/
	    if (pf->flags&F_SIGN) sep = "";	/* Sign WITH value */
	}
	else {	/* Which kind of alignement ? */
	    if (sep) BUF_AppendString(b, "</TD>") ;
	    if (pf->ff == 'A') tag = pf->len > nowrap ?
	        "<TD>" :
		"<TD NOWRAP>" ;
	    else tag = "<TD ALIGN=\"RIGHT\" NOWRAP>";
	    sep = tag ;
	}
	if (sep) BUF_AppendString(b, sep);	/* Not in very first */
	buf_addcol(b, l, strip, pf);
    }
    if (sep) BUF_AppendString(b, "</TD>") ;
    BUF_AppendString(b, "</TR>\n");
    p = b->buf;	/* To Return */
    free(b);
    return(p);
}

static PCHAR_FCT tbm_mapping[] = {
    tbm_ascii,		/* 'a' -- ascii   */
    tbm_csv,  		/* 'c' -- CSV     */
    tbm_par, 		/* 'p' -- parfile */
    tbm_latex, 		/* 't' -- latex	  */
    tbm_htx, 		/* 'h' -- htx	  */
    tbm_html, 		/* 'H' -- HTML	  */
} ;

int tbm_free(TABMAP  *M)
/*++++++++++++++++
.PURPOSE  Drop the MAP
.RETURNS  0 = OK
-----------------*/
{
    if (!M) return(0);
    free(M);
    return(0);
}

TABMAP *tbm_new(void *tab, char *option, char *list)
/*++++++++++++++++
.PURPOSE  Create a new Mapping Structure, for table Edition
.RETURNS  A new structure
.REMARKS  Option either lowercase, or single-char upper : H for HTML
	Includes * ** | || etc...
	List is either a full operation including ${...},
	or just a list of columns. The list of columns may
	use * (std set), ** (whole set)
-----------------*/
{
  COLINFO *colist = (COLINFO *)0;
  TABINFO *T ;
  TABMAP  *M ;
  char  mode ;
  char *text ;		/* Instruction */
  int imode ;
  char *p ;
  int  i=0;

    /* Is it from a FILINFO or a TABINFO ? */
    T = (TABINFO *)0 ;
    theFile = (FILINFO *)0 ;
    Wrecord.len = 0;
    if (((TABINFO *)tab)->signature) {
        T = (TABINFO *)tab ;
        if ((T->signature[0] == '\3') && (T->signature[1] == 'F')) {
	    theFile = (FILINFO *)tab ;
	    Wrecord.len = theFile->reclen;
	}
	if (theFile) T = theFile->tabform ;
    }
    if (!T) {
        fprintf(stderr, "****tbm_new: Can't interpret TABINFO at 0x%p\n", tab);
	return((TABMAP*)0) ;
    }
    mode = *option;
    /* V4.5: empty list or '*' equivalent to all columns */
    if (list) {
	for (p=list; *p == '*'; p++) ;
	if (!*p) list = (char *)0;
    }
    if ((option[0] == 'h') && (option[1] == 't') && (option[2] == 'm'))
	mode = 'H' ;
    imode = strloc(modes, mode) ;
    if (!modes[imode]) {
        fprintf(stderr,
    	    "#+++tbm_new: can't understand option \"%s\"\n", option);
	return((TABMAP *)0) ;
    }
    if (islower(*option)) while(islower(*option)) option++ ;
    else option++ ;
    if (list) {
	if (strchr(list, '$')) {	/* Standard mapping */
	    text = malloc(12+strlen(list));
	    sprintf(text, "#!form=%c\n%s", modes[imode], list);
	}
	else {
	    if ((colist = col_choice(T, list)))
		text = (*tbm_mapping[imode])(T, option, colist);
	    else text = (char *)0;
	}
    }
    else text = (*tbm_mapping[imode])(T, option, colist); /* Default Mapping */
    if (colist) free(colist);
    if (!text) return((TABMAP*)0) ;
    M = new(T, text);
    if (!M) {
	fprintf(stderr, "****tabmap: can't map (");
	for (p=text; *p && isprint(*p); p++) fputc(*p, stderr) ;
	fputc(')', stderr);
	if (*p) {
	    fprintf(stderr, ":\n   \"");
	    for (++p; *p; p++) {
		if (isprint(*p)) fputc(*p, stderr);
		else switch(*p) {
		  case '\n': fprintf(stderr, "\\n"); break;
		  case '\t': fprintf(stderr, "\\t"); break;
		  case '\v': fprintf(stderr, "\\v"); break;
		  case '\r': fprintf(stderr, "\\r"); break;
		  default: fprintf(stderr, "\\%03o", *p&0xff);
		}
	    }
	    fputc('"', stderr);
	}
	fputc('\n', stderr);
	free(text);
	return(M);
    }
    if (theFile) M->dfile = theFile->dfile ;
    free(text);

    /* Count the special options */
    if (option) for (p=option, i=0; isgraph(*p); p++) {
	if (*p == '|') M->bars += 1;
	if (*p == '_') M->flags |= F__MAP ;
	if (*p == '*') i++ ;
    }
    if (i > sizeof(flagstars)) i = sizeof(flagstars);
    if (i) M->flags |= flagstars[i-1] ;

    /* Prepare the Titles */
    make_title(M) ;

    return(M) ;
}

static CATINFO *getReadMe(char *arg, int *ofile)
/*++++++++++++++++
.PURPOSE  Interpret a catalog/table
.RETURNS  The CATINFO (ReadMe) ; ofile specifies the offset
-----------------*/
{
  char *tofree, *p;
  CATINFO *T;
  int floc=0;

    if ((tofree = strrchr(arg, '/'))) {
	tofree = p = malloc(strlen(arg)+strlen(readme_file)+4);
	strcpy(tofree, arg);
	/* Catalog may be separated by /./ */
	if ((*p == '.') || (p = strstr(tofree, "/./"))) {   /* Take LAST /./ */
	  char *f;
	    for (f=p; f; f=strstr(f+1, "./")) p=f;
	}
	else p = strrchr(tofree, '/')+1;
	floc = (p-tofree);	/* Length of catalog part (V5.5) */
	if (ofile) *ofile = floc;
	strcpy(p, "ReadMe");
    }
    p = tofree ? tofree : readme_file /*"ReadMe"*/;
    T = cad_new(p, fop);
    while ((!T) && tofree) {	/* V5.69: try to go down hierarchy */
	if (fop&FOP_v) fprintf(stderr, 
	    "#+++'%s' file not found: %s\n", readme_file, tofree);
	if ((p = strrchr(tofree, '/'))) *p = 0; else break;
	if ((p = strrchr(tofree, '/'))) p++;    else p = tofree;
	floc = (p-tofree);	/* Length of catalog part */
	strcpy(p, readme_file); /*strcpy(p, "/ReadMe");*/
	T = cad_new(tofree, fop);
    }
    if ((fop&FOP_v) && T) {  SECTION *ps; int i; FILINFO *pi;
	fprintf(stderr, "====Section Summary of File %s:\n", p);
    	for (ps=T->section[0]; ps; ps = ps->next)
    	    sec_prhead(ps, stderr);
	for (i=T->nfinfo, pi=T->afinfo; --i >= 0; pi++)
	    if (pi->reclen) printf("%-15s %5d %8d    %s\n",
	    pi->dfile, pi->reclen, pi->records, pi->text);
    }
    if (tofree) free(tofree);
    if (ofile) *ofile = floc;
    return(T);
}

#if TEST
#if 0
static int trim(char *buf)
/*++++++++++++++++
.PURPOSE  Remove trailing blanks
.RETURNS  Length
.REMARKS  Ending blanks / newlines replaced by zeroes
-----------------*/
{
  char *p ;
    for (p=buf+strlen(buf)-1; (p >= buf) && isspace(*p); p--) *p = 0 ;
    return(p-buf) ;
}
#endif

static int int_value(char *str)
/*++++++++++++++++
.PURPOSE  Check a correct number
.RETURNS  -1 if error
-----------------*/
{
  int i;
    for(i=0; isdigit(str[i]); i++) ;
    if (str[i]) return(-1);
    return(atoi(str));
}

/*==================================================================
		Test Program
 *==================================================================*/
int main (int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
{
  static char *tests[] = {
    "tex|*",
    "pS0*",
    "c|*",
    "a|",
    "a||",
    "html*|",
    (char *)0
  } ;
  TABMAP *tbm=0; TABINFO *tbf; FILINFO *pi; CATINFO *T; 
  int i, nfile, len, headone, printed, iscomment, datalines, floc, truncated;
  int add_blank_line=1;	/* V5.94: insert blank line between files, 
			   but not if comma-separated */
  FILE *fdata;
  char **a=0, *p=0, *arg=0, *tofree, *next_file;

    /* Initialisations */
    T = (CATINFO*)0;
    nfile = 0;
    pi = (FILINFO *)0;
    Wrecord.ff = 'W';
    Wrecord.rep = 1;
    Wrecord.unit  = "";
    Wrecord.label = "WholeRecord";
    Wrecord.text  = "the whole line of the catalog";

    Crecord.ff = 'I';		/* Edition of counter */
    Crecord.rep = 1;
    Crecord.len = 6;		/* Length of edition (should be changeable?) */
    Crecord.unit  = "";
    Crecord.label = "recno";
    Crecord.text  = "Counter of record in the table (from 1)";
    Crecord.flags |= F_USER;

    /* Get the Options, and Execute */
  go_fromOptions:
    while (--argc > 0) {
	arg = *++argv;
	if (*arg == '-') switch(arg[1]) {
	  case 'a':
	    if (strncmp(arg, "-asci", 5) == 0) { form = arg+1; continue; }
	    goto case_default;
	  case 'c':
	    if (strncmp(arg, "-csv", 4) == 0)  { form = arg+1; continue; }
	    goto case_default;
	  case 'p':
	    if (strncmp(arg, "-par", 4) == 0)  { form = arg+1; continue; }
	    goto case_default;
	  case 'b':
	    bopt = 1;		/* In -b, keep also columns separated */
	  case 'k':		/* Keep columns separated */
	    keep_cols = 1;	/* Don't merge columns */
	    continue;
	  case 'v':
	    if (strncmp(arg, "-ver", 4) == 0) {
		printf("tabmap (CDS) Version %s\n", VERSION);
		exit(0);
	    }
	    fop = FOP_v;
	    filutil(1);
	    continue;
	  case 'd':
	    issue_titles = 0;	/* No titles */
	    head_opt[0] = 0;	/* Remove the headers */
	    continue;
	  case 'H':	/* HTML */
	    form = arg+1; continue;
	  case 'h':	/* HELP */
	    if ((strncmp(arg, "-hel", 4) == 0) || (argc <= 1)) {  /* Help */
	        printf("%s%s", usage, help);
	        exit(0);
	    }
	    if (strncmp(arg, "-htx", 4) == 0) { form = arg+1; continue; }
	    if (strncmp(arg, "-htm", 4) == 0) {
		form = strdup(arg+1);
		form[0] = 'H';
		for (p=form+1; islower(*p); p++) ;
		for (i=1; *p; p++) form[i++] = *p;
		form[i] = 0;
		continue;
	    }
	    /* Specify the hearder characteristics */
	    --argc; ++argv;
	    head_opt = *argv;
	    continue;
	  case 'g':	/* -go (V5.76) */
	    if (strcmp(arg, "-go") == 0) {
		if (mapping) { free(mapping); mapping = 0; }
		if (tbm) { tbm_free(tbm); tbm = 0; }
		continue;
	    }
	    goto case_default;
	  case 'i':	/* Text to issue */
	    if (arg[2]) p = arg+2;
	    else if ((p = *++argv)) --argc;
	    else goto missing_Argument;
	    interpret_escape(p, 1); 
	    if (tbm && (tbm->mode == 'p')) putchar('\n');
	    printf("%s", p); 
	    continue;
	  case 'f':	/* Form type */
	    --argc;
	    if ((form    = *++argv)) continue;
	    goto missing_Argument;
	  case 'm':	/* Mapping type */
	    /* Accept max_lines -1 for All (V5.93) */
	    if (isdigit(arg[2])) { max_lines = atoi(arg+2); continue; }
	    if ((strcmp(arg, "-max") == 0) || (strcmp(arg, "-ml") == 0)) {
		--argc;
		if (*++argv) max_lines = int_value(*argv);
		else  goto missing_Argument;
		if(max_lines<0) {
		    /* Accept -ml - to indicate no truncation */
		    if(argv[0][0] == '-') max_lines = 0x7fffffff;
		    else goto numeric_Argument;
		}
		continue;
	    }
	  case 'e':
	    omapc = strchr(arg, 'c') != (char *)0;
	    --argc;
	    if (*++argv) {	/* Should terminate by \n */
		if (mapping) free(mapping);
		mapping = malloc(2+strlen(*argv)) ;
		strcpy(mapping, *argv);
		interpret_escape(mapping, omapc);
		p = mapping + strlen(mapping);
		*(p++) = '\n';
		*p = 0;
		continue;
	    }
	    goto missing_Argument;
	  case 'n':	/* NULL pattern	*/
	    if (strncmp(arg, "-note", 5) == 0) {    /* -note: issue the notes */
		P_opt |= P_note0;
		continue;
	    }
	    if ((p = strchr(arg, '='))) { null_value = p+1; continue; }
	    --argc;
	    if ((null_value = *++argv)) continue;
	    goto missing_Argument;
	  case 'l':
	    if (strcmp(arg, "-latex")==0) { form = "tex"; continue; }
	    goto case_default;
	  case 's':	/* SKIP or START records */
	    --argc;
	    if (*++argv) oskip = int_value(*argv);
	    else  goto missing_Argument;
	    if (arg[2] == 't') 	/* V5.83 */ oskip--;
	    if (oskip<0) goto numeric_Argument;
	    continue;
	  case 'r':
	    --argc;
	    if (strncmp(arg, "-recno", 6) == 0) { /* Equivalent to -start */
		if (*++argv) oskip = int_value(*argv);
		else  goto missing_Argument;
		oskip--;
		if (oskip<0) goto numeric_Argument;
		continue;
	    }
	    readme_file = *++argv;
	    if (!readme_file) {
		fprintf(stderr, "#***Missing readme_file name\n");
		exit(1);
	    }
	    continue;
	  case 't':
	    if ((arg[2]==0) && (argv[1])) { /* -t option: Use a template */
		tableform = *++argv; argc--;
		/* V5.68: if a '/' exists ==> use the corresponding catalog */
		if (strchr(tableform, '/')) {	
		    if (T) cad_kill(T);
		    if ((T = getReadMe(tableform, &floc)))  /* catalog/table */
		        tableform += floc;
		}
		continue;
	    }
	    if (strcmp(arg, "-tex") == 0) { form = arg+1; continue; }
	    if (strcmp(arg, "-txt") == 0) { form = "asci"; continue; }
	    if (strcmp(arg, "-tsv") == 0) { form = "csv"; continue; }
	    if (strcmp(arg, "-tail") == 0) { 	/* Idem +n */
		if (*++argv) otail = int_value(*argv);
		else  goto missing_Argument;
		if (otail<0) goto numeric_Argument;
		--argc ;
		continue;
	    }
	    tourop = 1;		     /* Show the possibilities */
	    continue;
	  case 0:                    /* "-" =  standard input  */
	    goto arg_isafile;
	  default:
	    if (isdigit(arg[1])) {	/* e.g. -10 (10 lines) */
		max_lines = int_value(arg+1);
		if (max_lines<0) goto numeric_Argument;
		continue;
	    }
	  case_default:
	    fprintf(stderr, "****Invalid argument: %s\n%s", arg, usage);
	    exit(1);
	  missing_Argument:
	    fprintf(stderr, "****Missing argument to %s\n%s", arg, usage);
	    exit(1);
	  numeric_Argument:
	    fprintf(stderr, "****Bad argument value to %s\n%s", arg, usage);
	    exit(1);
	}
	else if (arg[0] == '+') {
	    if (strncmp(arg, "+note", 5) == 0) {
		P_opt &= (P_note0|P_note1);
		P_opt |= P_note1;
		continue;
	    }
	    if (isdigit(arg[1])) { /* V5.2: tail */
	        otail = int_value(arg+1);
		/* if (otail>0) otail;	.. V5.62 */
		if (otail<0) goto numeric_Argument;
	        continue;
	    }
	}

      arg_isafile:
	/* I may have a list of instruction (i.e. -map missing) */
	if ((argc>1) && (!mapping) && argv[1]) {	/* Mod. V4.9 */
	    mapping = malloc(2+strlen(*argv)) ;
	    strcpy(mapping, *argv);
	    interpret_escape(mapping, omapc);
	    p = mapping + strlen(mapping);
	    *(p++) = '\n';
	    *p = 0;
	    continue;
	}

	/* I may have a file directory/file */
	floc = 0;	/* Position, in arg, of file name */
	if (!T) {
	    T = getReadMe(arg, &floc); /* catalog/table */
	    if (!T) {
		fprintf(stderr, 
		    "****tabmap: can't find ReadMe file for: %s\n", arg);
		exit(1);
	    }
	    theCat = T;
#if 0	/* in getReadMe */
	    tofree = strrchr(arg, '/');
	    if (tofree) {
		tofree = malloc(strlen(arg)+8);
		strcpy(tofree, arg);
		/* Catalog may be separated by /./ */
		if ((p = strstr(tofree, "./"))) {	/* Take LAST /./ */
		  char *f;
		    for (f=p; f; f=strstr(f+1, "./")) p=f;
		}
		else p = strrchr(tofree, '/')+1;
		floc = (p-tofree);	/* Length of catalog part (V5.5) */
		strcpy(p, "ReadMe");
	    }
	    p = tofree ? tofree : "ReadMe";
	    T = cad_new(p, fop);
	    if (!T) exit(1);
	    theCat = T;
            if (fop&FOP_v) {
		fprintf(stderr, "====Section Summary of File %s:\n", p);
    		for (ps=T->section[0]; ps; ps = ps->next)
    		    sec_prhead(ps, stderr);
		for (i=T->nfinfo, pi=T->afinfo; --i >= 0; pi++)
		    if (pi->reclen) printf("%-15s %5d %8d    %s\n",
		    pi->dfile, pi->reclen, pi->records, pi->text);
	    }
	    if (tofree) free(tofree);
#endif
	}

	/* Which file now? */
     Do_next_file:
	next_file = (char *)0;         /* V5.94: several files */
	tofree = (char *)0;
	if (tableform) {         /* Template specified as argument */
	    pi = cad_findfile(T, tableform);
	    if ((arg[0] == '-') && (arg[1] == 0))   /* stdin */
		fdata = stdin;
	    else if ((fdata = fzopen3(arg, oskip, 0)))  oskip = 0;
	    else if ((next_file = strchr(arg, ','))) { 
		/* V5.94: maybe several files separated by , */
		tofree = malloc(strlen(arg)+6); strcpy(tofree, arg);
		*strchr(tofree, ',') = 0; arg = ++next_file;
		if ((fdata = fzopen3(tofree, oskip, 0)))  oskip = 0;
	    }
	}
	else {                   /* I have a file specified */
	    /* ALL FILES indicates by . or nothing */
	    p = arg + floc;	/* V5.5: name of file */
	    /* if ((arg[1] == 0) && strchr("*.", arg[0])) [Replaced in V5.51] */
	    if ((p[0] == 0) || ((p[1] == 0) && ((*p=='.')||(*p=='*')))) {
		/* All files */
		argv = (char **)malloc((T->nfinfo+2)*sizeof(char *));
		argc = 0;
		argv[argc++] = "tabmap";
		for (i=T->nfinfo, pi=T->afinfo; --i>=0; pi++) {
		    if (!pi->tabform) continue;
		    argv[argc++] = p = malloc(floc+strlen(pi->dfile)+2);
		    if (floc) strcpy(p, arg);
		    strcpy(p+floc, pi->dfile);
		}
		argv[argc] = (char *)0;
		goto go_fromOptions;
	    }

	    if (fop&FOP_v) fprintf(stderr, "#...ready to (z)open: %s\n", arg);
	    fdata = fzopen3(arg, oskip, 0);
	    if ((!fdata) && (next_file = strchr(arg, ','))) { 
		/* V5.94: maybe several files separated by , */
		tofree = malloc(strlen(arg)+6); strcpy(tofree, arg);
		*strchr(tofree, ',') = 0; arg = ++next_file;
		fdata = fzopen3(tofree, oskip, 0);
	    }
	    if (!fdata) {
		p = tofree ? tofree : arg;
		if (!strchr(p, '.')) {
		    if (tofree) strcat(tofree, ".dat");
		    else tofree = strdup2(arg, ".dat");
	            fdata = fzopen3(tofree, oskip, 0);
		}
	    }
	}
	if (!fdata) {				/* No data file */
	    fprintf(stderr, "#****can't access file: %s\n",
		    tofree ? tofree : arg);
	    exit(1);
	}
	/*oskip = 0;*/
	if (tofree) free(tofree);
	if ((nfile>0) && add_blank_line) putchar('\n');
	if (verbop) {
	    tofree = fzinfo(fdata);
	    fprintf(stderr, "#...open(%s): %s\n", arg, 
		    tofree ? tofree : "(ok)");
	}
	tofree = 0; nfile++;

	/* Be sure the file is described */
	if (!tableform) {
	    if ((p = strstr(arg, "./"))) {
	        /* When the file is written as catalog/./file */
	        p += 2;
	        while (strstr(p, "./")) p = strstr(p, "./")+2;
	    }
	    /*else if ((p = strrchr(arg, '/'))) p++;*/
	    else p = arg+floc;
	    pi = cad_findfile(T, p);
	}
	if (!pi) {
	    fprintf(stderr, "****File '%s' not found!\n", p);
	    exit(1);
	}
	tbf = pi->tabform;
	if (! tbf) {
	    fprintf(stderr, "****File '%s' not described!\n", p);
	    exit(1);
	}

	/* Find the mapping instructions if not specified */
	if ((!form) && (!mapping)) {
	    if (tableform) { /* Assume asci if a table specified */
		form="a"; a = &form;
		mapping = "";
	    }
	    else if (!tourop) {
		fprintf(stderr, "#+++Missing -form, assuming -tour\n");
	        tourop = 1;
	        a = tests;
	    }
	    else {	/* -tour, but a not defined */
		a = tests;
	    }
	}
	else {
	    if (!mapping) mapping = "";
	    if (!form) form = "a";
	    a = &form;
	}
	len = pi->reclen + 2;
	/* Len is not known for generic files, e.g. subdir/ *.dat */
	if (len == 2) len = 3 + tbf->afmt[tbf->nfmt-1].b1;  /* V4.4 */
	if (len < 512) len = 512;
	tofree = malloc(len);
	while(*a) {
	    if (tbm) tbm_free(tbm);
	    tbm = tbm_new(pi, *a, mapping);
	    if (!tbm) exit(1);
	    tbm->dfile = arg;	/* V5.92: Full File Name */
	    if (fop&FOP_v) fprintf(stderr,
		"#...Edit file '%s' with form=%s\n%s\n#---(End of form)\n",
		arg, *a, tbm->text);
	    if (tourop) {
		printf("\n....Edit 10 rows of file '%s' with form=%s\n%s %s",
		        arg, *a, tbm->text, "(hit Return)");
		if (isatty(0)) i = getchar();
		else putchar('\n');
	    }
	    headone = issue_titles;
	    recno = printed = datalines = truncated = 0;
	    while(fgets(tofree, len, fdata)) {
		iscomment = recno < tbf->headlines
		         || tbf->headchar == tofree[0];
		/* V4.0: empty lines are IGNORED -- except when titles asked,
		 * or in comment... */
		if (tofree[0] == '\n') {
		    if (otail|iscomment) ++recno;		/* V5.62 */
		    if (issue_titles) printf("#%s", tofree);
		    continue;
		}
		++recno;
		if (otail && (recno < otail))   /* V5.2 */
		    continue;
		/* V4.81: Issue title AFTER comment(s) */
		if (headone) {		/* headone=1: MUST issue title */
		    if (iscomment) headone = 2;
		    else headone = 1;
		}
		if (headone == 1) {
		    if (head_opt[0] == 'H') {
			tbm_prhead(tbm, stdout, P_opt);
		    }
	            if (tbm->prefix) fprintf(stdout, "%s", tbm->prefix);
		    headone = 0;
	        }
		if (max_lines && (printed>=max_lines)) {    /* V4.6 */
		    if(verbop) fprintf(stderr, "#+++MAX (%d>=%d) reached\n",
			    printed, max_lines);
		    truncated = 1;
		    break;		/* STOPPED due to max_lines */
		}
		if (iscomment) {
		    if (issue_titles) printf("#%s", tofree);
		    /* V4.91: Comment may be longer than specified length */
		    while (!strchr(tofree, '\n')) {
			if (fgets(tofree, len, fdata) && issue_titles)
			    printf("%s", tofree);
		    }
		}
		else {
		    datalines++;
		    if (datalines <= oskip)	/* Added V5.1 */
			continue;
		    tbm_pr1(tbm, stdout, tofree);
		    printed++;
		}
		if (tourop && (printed >= 10)) break;
	    }
	    if (issue_titles && tbm->postfix)
		fprintf(stdout, "%s", tbm->postfix);
	    if (truncated) { 
		sprintf(tofree, "Output truncated to %d records", max_lines);
		print_message(tbm, stdout, tofree);
	    }
	    if (P_opt&P_note1)
		tbm_prnotes(tbm, stdout, P_opt);
	    if (tourop) a++;
	    else break;
	}
	fflush(stdout);
	if (tofree) free(tofree); tofree = (char *)0;
	if (verbop) fprintf(stderr, "#...ready to (z)close %s\n", arg);
	i = fzclose(fdata);
	if (verbop) fprintf(stderr, "#...[(z)close status]: %d\n", i);
	oskip = 0;
	if(next_file) { add_blank_line=0; goto Do_next_file; }
	else add_blank_line = 1;
    }
    if (nfile == 0) {
	argv[0] = "tabmap";
	argv[2] = (char *)0;
	argc = 2;
	if (tableform) {
	    argv[1] = "-";
	    goto go_fromOptions;
	}
	else if (tourop) { 
	    if (!T) T = cad_new(strdup2("./", readme_file)/*"./ReadMe"*/, fop);
	    if (T) pi=T->afinfo;
	    argv = malloc(3*sizeof(char *));
	    if (pi) {
		for (i=0; i<T->nfinfo; i++) {
		    if (strstr(pi->dfile, ".dat")) break;
		    pi++;
		}
		argv[1] = pi->dfile;
		if (i < T->nfinfo) goto go_fromOptions;
	    }
	}
	fprintf(stderr, "#+++Nothing done (missing table to process?)\n");
	exit_status = 2;
    }
    if (conclude && arg) printf(conclude, arg);
    /* cad_kill(T); */
    exit(exit_status);
}
#endif
